from uuid import uuid4

from deepdiff import DeepDiff

from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.ports.port import Port
from vellum.workflows.references.vellum_secret import VellumSecretReference
from vellum_ee.workflows.display.base import WorkflowInputsDisplay
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.types import NodeOutputDisplay


class Inputs(BaseInputs):
    input: str


def test_serialize_node__basic(serialize_node):
    class BasicGenericNode(BaseNode):
        pass

    serialized_node = serialize_node(BasicGenericNode)
    assert not DeepDiff(
        {
            "id": "7fc62d68-0a27-4506-b4ea-e40bc116c0c5",
            "label": "Basic Generic Node",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "BasicGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "trigger": {"id": "454606d2-204a-4b29-b0dd-59233c9e5a0e", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "4173b79f-ebdd-4026-8046-01d20dcf2b13",
                    "name": "default",
                    "type": "DEFAULT",
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__if(serialize_node):
    class IfGenericNode(BaseNode):

        class Ports(BaseNode.Ports):
            if_branch = Port.on_if(Inputs.input.equals("hello"))

    input_id = uuid4()
    serialized_node = serialize_node(
        node_class=IfGenericNode, global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=input_id)}
    )

    assert not DeepDiff(
        {
            "id": "108e36a6-a773-4bc7-b6c4-38997e3690c5",
            "label": "If Generic Node",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "IfGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "trigger": {"id": "6769e811-fc56-4bab-9ebb-4b4b4b5ef023", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "cec9e02a-4097-4e0a-b62d-60da4f17db67",
                    "type": "IF",
                    "name": "if_branch",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {
                            "type": "WORKFLOW_INPUT",
                            "input_variable_id": str(input_id),
                        },
                        "operator": "=",
                        "rhs": {
                            "type": "CONSTANT_VALUE",
                            "value": {
                                "type": "STRING",
                                "value": "hello",
                            },
                        },
                    },
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__if_else(serialize_node):
    class IfElseGenericNode(BaseNode):

        class Ports(BaseNode.Ports):
            if_branch = Port.on_if(Inputs.input.equals("hello"))
            else_branch = Port.on_else()

    input_id = uuid4()
    serialized_node = serialize_node(
        node_class=IfElseGenericNode, global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=input_id)}
    )

    assert not DeepDiff(
        {
            "id": "6b16a6e3-cbe4-4672-b2bb-31ef930b0d97",
            "label": "If Else Generic Node",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "IfElseGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "trigger": {"id": "b4a3e968-0f88-457c-b91e-e6c490b669de", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "b3394650-f12e-443c-8abe-675bfb632cf5",
                    "type": "IF",
                    "name": "if_branch",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {
                            "type": "WORKFLOW_INPUT",
                            "input_variable_id": str(input_id),
                        },
                        "operator": "=",
                        "rhs": {
                            "type": "CONSTANT_VALUE",
                            "value": {
                                "type": "STRING",
                                "value": "hello",
                            },
                        },
                    },
                },
                {
                    "id": "7c213f16-ae88-4080-8b44-2084d4280c3f",
                    "type": "ELSE",
                    "name": "else_branch",
                    "expression": None,
                },
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
    )


def test_serialize_node__if_elif_else(serialize_node):
    class IfElifElseGenericNode(BaseNode):

        class Ports(BaseNode.Ports):
            if_branch = Port.on_if(Inputs.input.equals("hello"))
            elif_branch = Port.on_elif(Inputs.input.equals("world"))
            else_branch = Port.on_else()

    input_id = uuid4()
    serialized_node = serialize_node(
        node_class=IfElifElseGenericNode,
        global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=input_id)},
    )

    assert not DeepDiff(
        {
            "id": "da8860e8-e1ea-4e57-977b-7a8ef43415ee",
            "label": "If Elif Else Generic Node",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "IfElifElseGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "trigger": {"id": "1b86a845-df3a-4544-ae5b-a23887e735b1", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "d0fa9794-bd46-4238-b61f-6b03ba7053ba",
                    "type": "IF",
                    "name": "if_branch",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {
                            "type": "WORKFLOW_INPUT",
                            "input_variable_id": str(input_id),
                        },
                        "operator": "=",
                        "rhs": {
                            "type": "CONSTANT_VALUE",
                            "value": {
                                "type": "STRING",
                                "value": "hello",
                            },
                        },
                    },
                },
                {
                    "id": "5b4afa12-5c0d-4e6a-b304-d56baf4738dc",
                    "type": "ELIF",
                    "name": "elif_branch",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {
                            "type": "WORKFLOW_INPUT",
                            "input_variable_id": str(input_id),
                        },
                        "operator": "=",
                        "rhs": {
                            "type": "CONSTANT_VALUE",
                            "value": {
                                "type": "STRING",
                                "value": "world",
                            },
                        },
                    },
                },
                {
                    "id": "055fc9f8-3510-47d2-8dcf-d7bd0ea15fd6",
                    "type": "ELSE",
                    "expression": None,
                    "name": "else_branch",
                },
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
    )


def test_serialize_node__node_output_reference(serialize_node):
    class NodeWithOutput(BaseNode):

        class Outputs(BaseNode.Outputs):
            output = Inputs.input

    class NodeWithOutputDisplay(BaseNodeDisplay[NodeWithOutput]):
        pass

    class GenericNodeReferencingOutput(BaseNode):

        class Ports(BaseNode.Ports):
            if_branch = Port.on_if(NodeWithOutput.Outputs.output.equals("hello"))

    workflow_input_id = uuid4()
    node_output_id = uuid4()
    serialized_node = serialize_node(
        node_class=GenericNodeReferencingOutput,
        global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=workflow_input_id)},
        global_node_displays={NodeWithOutput: NodeWithOutputDisplay()},
        global_node_output_displays={
            NodeWithOutput.Outputs.output: NodeOutputDisplay(id=node_output_id, name="output")
        },
    )

    assert not DeepDiff(
        {
            "id": "e5387a2f-ccff-43b2-8b29-8f4ba9beb01e",
            "label": "Generic Node Referencing Output",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "definition": {
                "name": "GenericNodeReferencingOutput",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "trigger": {"id": "c43b6517-102a-44bc-83c8-37d42861d3e6", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "6258cb70-4cfc-4bc8-a26c-df14a50bc72a",
                    "type": "IF",
                    "name": "if_branch",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {
                            "type": "NODE_OUTPUT",
                            "node_id": "242ad818-12ec-449f-87b4-d2bc513e2148",
                            "node_output_id": str(node_output_id),
                        },
                        "operator": "=",
                        "rhs": {
                            "type": "CONSTANT_VALUE",
                            "value": {
                                "type": "STRING",
                                "value": "hello",
                            },
                        },
                    },
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__vellum_secret_reference(serialize_node):
    class GenericNodeReferencingSecret(BaseNode):

        class Ports(BaseNode.Ports):
            if_branch = Port.on_if(VellumSecretReference(name="hello").equals("hello"))

    workflow_input_id = uuid4()
    serialized_node = serialize_node(
        node_class=GenericNodeReferencingSecret,
        global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=workflow_input_id)},
    )

    assert not DeepDiff(
        {
            "id": "b1e770cb-31dd-461f-9041-8470c9f3e098",
            "label": "Generic Node Referencing Secret",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "definition": {
                "name": "GenericNodeReferencingSecret",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "trigger": {"id": "96f50af1-8178-47b8-b8c0-97ca10e62845", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "aaad042f-02ce-400a-aad5-c9064925937b",
                    "type": "IF",
                    "name": "if_branch",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {"type": "VELLUM_SECRET", "vellum_secret_name": "hello"},
                        "operator": "=",
                        "rhs": {
                            "type": "CONSTANT_VALUE",
                            "value": {
                                "type": "STRING",
                                "value": "hello",
                            },
                        },
                    },
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__execution_count_reference(serialize_node):
    class NodeWithExecutions(BaseNode):
        pass

    class NodeWithExecutionsDisplay(BaseNodeDisplay[NodeWithExecutions]):
        pass

    class GenericNodeReferencingExecutions(BaseNode):

        class Ports(BaseNode.Ports):
            if_branch = Port.on_if(NodeWithExecutions.Execution.count.equals(5))

    workflow_input_id = uuid4()
    serialized_node = serialize_node(
        node_class=GenericNodeReferencingExecutions,
        global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=workflow_input_id)},
        global_node_displays={NodeWithExecutions: NodeWithExecutionsDisplay()},
    )

    assert not DeepDiff(
        {
            "id": "5e6512c2-7a97-4b44-abed-64cee4388df5",
            "label": "Generic Node Referencing Executions",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "definition": {
                "name": "GenericNodeReferencingExecutions",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "trigger": {"id": "4a120ba5-2384-4854-a9fb-b9fe68bcd52a", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "575af12c-c700-438c-9123-198cca866535",
                    "type": "IF",
                    "name": "if_branch",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {
                            "type": "EXECUTION_COUNTER",
                            "node_id": "075e9647-b657-4a6e-ab21-c3a7f1a68ada",
                        },
                        "operator": "=",
                        "rhs": {
                            "type": "CONSTANT_VALUE",
                            "value": {
                                "type": "NUMBER",
                                "value": 5.0,
                            },
                        },
                    },
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__null(serialize_node):
    class NullGenericNode(BaseNode):

        class Ports(BaseNode.Ports):
            if_branch = Port.on_if(Inputs.input.is_null())

    input_id = uuid4()
    serialized_node = serialize_node(
        node_class=NullGenericNode, global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=input_id)}
    )

    assert not DeepDiff(
        {
            "id": "05d96ec5-4a05-43ca-a1c2-e92a49ac9520",
            "label": "Null Generic Node",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "NullGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "trigger": {"id": "bc28e72e-b75b-487c-8b8f-345c062d848d", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "b0b5f49f-7d02-4b45-9c8b-b4b5440a0e8e",
                    "type": "IF",
                    "name": "if_branch",
                    "expression": {
                        "type": "UNARY_EXPRESSION",
                        "lhs": {
                            "type": "WORKFLOW_INPUT",
                            "input_variable_id": str(input_id),
                        },
                        "operator": "null",
                    },
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__between(serialize_node):
    class IntegerInputs(BaseInputs):
        input: int

    class BetweenGenericNode(BaseNode):

        class Ports(BaseNode.Ports):
            if_branch = Port.on_if(IntegerInputs.input.between(1, 10))

    input_id = uuid4()
    serialized_node = serialize_node(
        node_class=BetweenGenericNode,
        global_workflow_input_displays={IntegerInputs.input: WorkflowInputsDisplay(id=input_id)},
    )

    assert not DeepDiff(
        {
            "id": "f32ab40b-7de2-4dc4-b5ef-41a753bf8bb4",
            "label": "Between Generic Node",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "BetweenGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "trigger": {"id": "99c92fd8-66af-4910-8f1c-773fa4c7fc42", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "9894ea65-ecbd-4d31-9ede-b65debebc908",
                    "type": "IF",
                    "name": "if_branch",
                    "expression": {
                        "type": "TERNARY_EXPRESSION",
                        "base": {
                            "type": "WORKFLOW_INPUT",
                            "input_variable_id": str(input_id),
                        },
                        "operator": "between",
                        "lhs": {
                            "type": "CONSTANT_VALUE",
                            "value": {
                                "type": "NUMBER",
                                "value": 1.0,
                            },
                        },
                        "rhs": {
                            "type": "CONSTANT_VALUE",
                            "value": {
                                "type": "NUMBER",
                                "value": 10.0,
                            },
                        },
                    },
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__or(serialize_node):
    class OrGenericNode(BaseNode):

        class Ports(BaseNode.Ports):
            if_branch = Port.on_if(Inputs.input.equals("hello") | Inputs.input.equals("world"))

    input_id = uuid4()
    serialized_node = serialize_node(
        node_class=OrGenericNode, global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=input_id)}
    )

    assert not DeepDiff(
        {
            "id": "8971f041-6b61-4d02-b02b-9d221f224ea2",
            "label": "Or Generic Node",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "OrGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "trigger": {"id": "69c1fb9e-2b89-46f2-a1b0-c7bd2e55d812", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "0306137d-1af9-430e-8582-08054d1ed350",
                    "type": "IF",
                    "name": "if_branch",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {
                            "type": "BINARY_EXPRESSION",
                            "lhs": {
                                "type": "WORKFLOW_INPUT",
                                "input_variable_id": str(input_id),
                            },
                            "operator": "=",
                            "rhs": {
                                "type": "CONSTANT_VALUE",
                                "value": {
                                    "type": "STRING",
                                    "value": "hello",
                                },
                            },
                        },
                        "operator": "or",
                        "rhs": {
                            "type": "BINARY_EXPRESSION",
                            "lhs": {
                                "type": "WORKFLOW_INPUT",
                                "input_variable_id": str(input_id),
                            },
                            "operator": "=",
                            "rhs": {
                                "type": "CONSTANT_VALUE",
                                "value": {
                                    "type": "STRING",
                                    "value": "world",
                                },
                            },
                        },
                    },
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__and_then_or(serialize_node):
    class AndThenOrGenericNode(BaseNode):

        class Ports(BaseNode.Ports):
            if_branch = Port.on_if(
                Inputs.input.equals("hello") & Inputs.input.equals("then") | Inputs.input.equals("world")
            )

    input_id = uuid4()
    serialized_node = serialize_node(
        node_class=AndThenOrGenericNode,
        global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=input_id)},
    )

    assert not DeepDiff(
        {
            "id": "ce9daa70-0e06-4ee4-b611-cd5702e92eed",
            "label": "And Then Or Generic Node",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "AndThenOrGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "trigger": {"id": "eddecfb5-4bdf-4193-b8c2-bc7bf1bd96b1", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "c87cfe7b-a6a9-43eb-a890-4785ee1783e7",
                    "type": "IF",
                    "name": "if_branch",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {
                            "type": "BINARY_EXPRESSION",
                            "lhs": {
                                "type": "BINARY_EXPRESSION",
                                "lhs": {
                                    "type": "WORKFLOW_INPUT",
                                    "input_variable_id": str(input_id),
                                },
                                "operator": "=",
                                "rhs": {
                                    "type": "CONSTANT_VALUE",
                                    "value": {
                                        "type": "STRING",
                                        "value": "hello",
                                    },
                                },
                            },
                            "operator": "and",
                            "rhs": {
                                "type": "BINARY_EXPRESSION",
                                "lhs": {
                                    "type": "WORKFLOW_INPUT",
                                    "input_variable_id": str(input_id),
                                },
                                "operator": "=",
                                "rhs": {
                                    "type": "CONSTANT_VALUE",
                                    "value": {
                                        "type": "STRING",
                                        "value": "then",
                                    },
                                },
                            },
                        },
                        "operator": "or",
                        "rhs": {
                            "type": "BINARY_EXPRESSION",
                            "lhs": {
                                "type": "WORKFLOW_INPUT",
                                "input_variable_id": str(input_id),
                            },
                            "operator": "=",
                            "rhs": {
                                "type": "CONSTANT_VALUE",
                                "value": {
                                    "type": "STRING",
                                    "value": "world",
                                },
                            },
                        },
                    },
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__parenthesized_and_then_or(serialize_node):
    class ParenthesizedAndThenOrGenericNode(BaseNode):

        class Ports(BaseNode.Ports):
            if_branch = Port.on_if(
                Inputs.input.equals("hello") & (Inputs.input.equals("then") | Inputs.input.equals("world"))
            )

    input_id = uuid4()
    serialized_node = serialize_node(
        node_class=ParenthesizedAndThenOrGenericNode,
        global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=input_id)},
    )

    assert not DeepDiff(
        {
            "id": "c311e689-f06c-47a9-88dc-ca2e1f0ae455",
            "label": "Parenthesized And Then Or Generic Node",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "ParenthesizedAndThenOrGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "trigger": {"id": "3d9cd609-7801-4ce5-b70c-2afcc88d8af3", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "b58e5bcb-cfcd-4249-b80d-447dc4b096f0",
                    "type": "IF",
                    "name": "if_branch",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {
                            "type": "BINARY_EXPRESSION",
                            "lhs": {
                                "type": "WORKFLOW_INPUT",
                                "input_variable_id": str(input_id),
                            },
                            "operator": "=",
                            "rhs": {
                                "type": "CONSTANT_VALUE",
                                "value": {
                                    "type": "STRING",
                                    "value": "hello",
                                },
                            },
                        },
                        "operator": "and",
                        "rhs": {
                            "type": "BINARY_EXPRESSION",
                            "lhs": {
                                "type": "BINARY_EXPRESSION",
                                "lhs": {
                                    "type": "WORKFLOW_INPUT",
                                    "input_variable_id": str(input_id),
                                },
                                "operator": "=",
                                "rhs": {
                                    "type": "CONSTANT_VALUE",
                                    "value": {
                                        "type": "STRING",
                                        "value": "then",
                                    },
                                },
                            },
                            "operator": "or",
                            "rhs": {
                                "type": "BINARY_EXPRESSION",
                                "lhs": {
                                    "type": "WORKFLOW_INPUT",
                                    "input_variable_id": str(input_id),
                                },
                                "operator": "=",
                                "rhs": {
                                    "type": "CONSTANT_VALUE",
                                    "value": {
                                        "type": "STRING",
                                        "value": "world",
                                    },
                                },
                            },
                        },
                    },
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__or_then_and(serialize_node):
    class OrThenAndGenericNode(BaseNode):

        class Ports(BaseNode.Ports):
            if_branch = Port.on_if(
                Inputs.input.equals("hello") | Inputs.input.equals("then") & Inputs.input.equals("world")
            )

    input_id = uuid4()
    serialized_node = serialize_node(
        node_class=OrThenAndGenericNode,
        global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=input_id)},
    )

    assert not DeepDiff(
        {
            "id": "4394a83e-897c-4fe1-a76c-6e8853a619a9",
            "label": "Or Then And Generic Node",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "OrThenAndGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "trigger": {"id": "ef0b522b-f9c0-4767-b7ea-f0fb15b213ed", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "2770f76e-f20e-40c6-86a0-8efbc08a8265",
                    "type": "IF",
                    "name": "if_branch",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {
                            "type": "BINARY_EXPRESSION",
                            "lhs": {
                                "type": "WORKFLOW_INPUT",
                                "input_variable_id": str(input_id),
                            },
                            "operator": "=",
                            "rhs": {
                                "type": "CONSTANT_VALUE",
                                "value": {
                                    "type": "STRING",
                                    "value": "hello",
                                },
                            },
                        },
                        "operator": "or",
                        "rhs": {
                            "type": "BINARY_EXPRESSION",
                            "lhs": {
                                "type": "BINARY_EXPRESSION",
                                "lhs": {
                                    "type": "WORKFLOW_INPUT",
                                    "input_variable_id": str(input_id),
                                },
                                "operator": "=",
                                "rhs": {
                                    "type": "CONSTANT_VALUE",
                                    "value": {
                                        "type": "STRING",
                                        "value": "then",
                                    },
                                },
                            },
                            "operator": "and",
                            "rhs": {
                                "type": "BINARY_EXPRESSION",
                                "lhs": {
                                    "type": "WORKFLOW_INPUT",
                                    "input_variable_id": str(input_id),
                                },
                                "operator": "=",
                                "rhs": {
                                    "type": "CONSTANT_VALUE",
                                    "value": {
                                        "type": "STRING",
                                        "value": "world",
                                    },
                                },
                            },
                        },
                    },
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__parse_json(serialize_node):

    class ParseJsonGenericNode(BaseNode):

        class Ports(BaseNode.Ports):
            if_branch = Port.on_if(Inputs.input.parse_json().equals({"key": "value"}))

    input_id = uuid4()
    serialized_node = serialize_node(
        node_class=ParseJsonGenericNode,
        global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=input_id)},
    )

    assert not DeepDiff(
        {
            "id": "60a14037-a8cf-47ad-b34b-47ec9dbf9715",
            "label": "Parse Json Generic Node",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "ParseJsonGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_ports_serialization",
                ],
            },
            "trigger": {"id": "2d711e0e-7b50-4eaf-ad2c-f6d7441b75f0", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "9eedb158-20a0-456e-b436-d9a7460b791e",
                    "type": "IF",
                    "name": "if_branch",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {
                            "type": "UNARY_EXPRESSION",
                            "lhs": {
                                "type": "WORKFLOW_INPUT",
                                "input_variable_id": str(input_id),
                            },
                            "operator": "parseJson",
                        },
                        "operator": "=",
                        "rhs": {
                            "type": "CONSTANT_VALUE",
                            "value": {
                                "type": "JSON",
                                "value": {"key": "value"},
                            },
                        },
                    },
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )



from vellum.utils.files.exceptions import FileNotFoundError, FileRetrievalError, InvalidFileSourceError, VellumFileError
from vellum.utils.files.mixin import VellumFileMixin
from vellum.utils.files.read import read_vellum_file
from vellum.utils.files.stream import stream_vellum_file
from vellum.utils.files.types import VellumFileTypes
from vellum.utils.files.upload import upload_vellum_file

__all__ = [
    "read_vellum_file",
    "stream_vellum_file",
    "upload_vellum_file",
    "VellumFileTypes",
    "VellumFileMixin",
    "VellumFileError",
    "InvalidFileSourceError",
    "FileRetrievalError",
    "FileNotFoundError",
]



from vellum.workflows.triggers.base import BaseTrigger
from vellum.workflows.triggers.integration import IntegrationTrigger
from vellum.workflows.triggers.manual import ManualTrigger
from vellum.workflows.triggers.schedule import ScheduleTrigger

__all__ = ["BaseTrigger", "IntegrationTrigger", "ManualTrigger", "ScheduleTrigger"]



"""File extension inference utilities."""

import mimetypes
import os
from typing import Optional

from vellum.utils.files.constants import EXTENSION_OVERRIDES, MIME_TYPE_TO_EXTENSION


def ensure_filename_with_extension(filename: Optional[str], mime_type: str) -> str:
    """
    Ensure the filename has an appropriate extension, infering one based on the provided MIME type if necessary.

    Args:
        filename: Optional filename provided by the user
        mime_type: The MIME type of the file (e.g., "application/pdf", "image/png"). This'll be used to infer the
            extension if the filename lacks one.

    Returns:
        A filename with an appropriate extension

    Examples:
        >>> ensure_filename_with_extension(None, "application/pdf")
        'file.pdf'
        >>> ensure_filename_with_extension("document", "application/pdf")
        'document.pdf'
        >>> ensure_filename_with_extension("document.pdf", "application/pdf")
        'document.pdf'
        >>> ensure_filename_with_extension("document.pdf", "image/jpeg")
        'document.pdf'  # Prioritizes existing extension
    """
    # Strip parameters from MIME type (e.g., "text/html; charset=utf-8" -> "text/html")
    mime_type = mime_type.split(";")[0].strip()

    # If no filename provided, start with "file"
    if not filename:
        base_name = "file"
        has_extension = False
    else:
        base_name = filename
        # Check if filename already has an extension
        _, ext = os.path.splitext(filename)
        has_extension = bool(ext)

    # If the provided filename already has an extension, keep it
    if filename and has_extension:
        return filename

    # Otherwise, infer extension from MIME type
    extension = mimetypes.guess_extension(mime_type)

    # Handle cases where mimetypes returns None or undesirable extensions
    if not extension:
        extension = MIME_TYPE_TO_EXTENSION.get(mime_type, ".bin")

    # Some MIME types have multiple possible extensions; prefer the most common ones
    extension = EXTENSION_OVERRIDES.get(extension, extension)

    return f"{base_name}{extension}"



from .constants import APIRequestMethod, AuthorizationType
from .errors.types import WorkflowErrorCode
from .exceptions import NodeException
from .inputs import BaseInputs, DatasetRow
from .nodes import (
    APINode,
    BaseInlinePromptNode,
    BaseNode,
    BasePromptDeploymentNode,
    BaseSearchNode,
    CodeExecutionNode,
    ConditionalNode,
    ErrorNode,
    FinalOutputNode,
    GuardrailNode,
    InlinePromptNode,
    InlineSubworkflowNode,
    MapNode,
    NoteNode,
    PromptDeploymentNode,
    RetryNode,
    SearchNode,
    SubworkflowDeploymentNode,
    TemplatingNode,
    TryNode,
    WebSearchNode,
)
from .nodes.displayable.tool_calling_node import ToolCallingNode
from .ports import Port
from .references.environment_variable import EnvironmentVariableReference
from .references.lazy import LazyReference
from .runner import WorkflowRunner
from .sandbox import WorkflowSandboxRunner
from .state.base import BaseState
from .triggers import IntegrationTrigger, ScheduleTrigger
from .types.core import Json, MergeBehavior
from .types.definition import DeploymentDefinition, MCPServer, VellumIntegrationToolDefinition
from .workflows import BaseWorkflow

__all__ = [
    "APINode",
    "APIRequestMethod",
    "AuthorizationType",
    "BaseInlinePromptNode",
    "BaseInputs",
    "BaseNode",
    "BasePromptDeploymentNode",
    "BaseSearchNode",
    "BaseState",
    "BaseWorkflow",
    "CodeExecutionNode",
    "ConditionalNode",
    "DatasetRow",
    "DeploymentDefinition",
    "EnvironmentVariableReference",
    "ErrorNode",
    "FinalOutputNode",
    "GuardrailNode",
    "InlinePromptNode",
    "InlineSubworkflowNode",
    "IntegrationTrigger",
    "Json",
    "LazyReference",
    "MapNode",
    "MCPServer",
    "MergeBehavior",
    "NodeException",
    "NoteNode",
    "Port",
    "PromptDeploymentNode",
    "RetryNode",
    "ScheduleTrigger",
    "SearchNode",
    "SubworkflowDeploymentNode",
    "TemplatingNode",
    "ToolCallingNode",
    "TryNode",
    "VellumIntegrationToolDefinition",
    "WebSearchNode",
    "WorkflowErrorCode",
    "WorkflowRunner",
    "WorkflowSandboxRunner",
]



# This file was auto-generated by Fern from our API Definition.

import typing

IntegrationName = typing.Union[
    typing.Literal[
        "SLACK",
        "NOTION",
        "GOOGLE",
        "CALENDLY",
        "CANVA",
        "CLICKUP",
        "CODA",
        "HUBSPOT",
        "INTERCOM",
        "LINEAR",
        "LINKUP",
        "LISTENNOTES",
        "LMNT",
        "LINKEDIN",
        "MAILCHIMP",
        "MEM0",
        "NEON",
        "GITHUB",
        "GOOGLE_SHEETS",
        "GOOGLE_CALENDAR",
        "GOOGLE_DRIVE",
        "GMAIL",
        "ACCULYNX",
        "AFFINITY",
        "AGENCYZOOM",
        "AHREFS",
        "AIRTABLE",
        "APOLLO",
        "ASANA",
        "ATLASSIAN",
        "BITBUCKET",
        "BOX",
        "BREVO",
        "BROWSERBASE_TOOL",
        "CAL",
        "CANVAS",
        "ELEVENLABS",
        "EXA",
        "GAMMA",
        "GITLAB",
        "FIRECRAWL",
        "FIGMA",
        "FIREFLIES",
        "GOOGLE_MAPS",
        "HEYGEN",
        "JIRA",
        "JUNGLESCOUT",
        "KLAVIYO",
        "PAGERDUTY",
        "PARSERA",
        "PEOPLEDATALABS",
        "PERPLEXITY",
        "POSTHOG",
        "PRODUCTBOARD",
        "REDDIT",
        "SEMRUSH",
        "SEMANTICSCHOLAR",
        "SENDGRID",
        "SERPAPI",
        "SHARE_POINT",
        "SHORTCUT",
        "STRIPE",
        "SUPABASE",
        "TAVILY",
        "TELEGRAM",
        "WEBFLOW",
        "YOUSEARCH",
        "ZENDESK",
        "ZENROWS",
        "DROPBOX",
        "EVENTBRITE",
        "CONFLUENCE",
        "COINBASE",
        "DISCORD",
        "DOCUSIGN",
        "TRELLO",
    ],
    typing.Any,
]



import io
import os
import sys
import traceback
from typing import Any, Optional, Tuple, Union

from vellum.workflows.constants import undefined
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.nodes.utils import cast_to_output_type, wrap_inputs_for_backward_compatibility
from vellum.workflows.state.context import WorkflowContext
from vellum.workflows.types.core import EntityInputsInterface


def read_file_from_path(
    node_filepath: str, script_filepath: str, context: Optional[WorkflowContext] = None
) -> Union[str, None]:
    node_filepath_dir = os.path.dirname(node_filepath)
    full_filepath = os.path.join(node_filepath_dir, script_filepath)

    # If dynamic file loader is present, try and read the code from there
    if context and context.generated_files:
        # Strip out namespace
        normalized_path = os.path.normpath(full_filepath)
        stripped_node_filepath = "/".join(normalized_path.split("/")[1:])

        code = context.generated_files.get(stripped_node_filepath, None)
        if code is not None:
            return code

    # Default logic for reading from filesystem
    try:
        with open(full_filepath) as file:
            return file.read()
    except (FileNotFoundError, IsADirectoryError):
        return None


def run_code_inline(
    code: str,
    inputs: EntityInputsInterface,
    output_type: Any,
    filepath: str,
) -> Tuple[str, Any]:
    log_buffer = io.StringIO()

    def _inline_print(*args: Any, **kwargs: Any) -> None:
        str_args = [str(arg) for arg in args]
        print_line = f"{' '.join(str_args)}\n"
        log_buffer.write(print_line)

    wrapped_inputs = wrap_inputs_for_backward_compatibility(inputs)
    exec_globals = {
        "__arg__inputs": wrapped_inputs,
        "__arg__out": None,
        "print": _inline_print,
    }
    run_args = [f"{name}=__arg__inputs['{name}']" for name, value in inputs.items() if value is not undefined]
    execution_code = f"""\
{code}

__arg__out = main({", ".join(run_args)})
"""
    try:
        compiled_code = compile(execution_code, filepath, "exec")
        exec(compiled_code, exec_globals)
    except Exception as e:
        lines = code.splitlines()
        _, _, tb = sys.exc_info()
        tb_generator = traceback.walk_tb(tb)
        stack = traceback.StackSummary.extract(tb_generator)

        # Filter stack to only include frames from the user's code file, and omit the first one
        filtered_stack = traceback.StackSummary.from_list([frame for frame in stack if frame.filename == filepath][1:])
        for frame in filtered_stack:
            if not frame.line and frame.lineno and frame.lineno <= len(lines):
                # Mypy doesn't like us setting private attributes
                frame._line = lines[frame.lineno - 1]  # type: ignore[attr-defined]

        error_message = f"""\
Traceback (most recent call last):
{''.join(filtered_stack.format())}
{e.__class__.__name__}: {e}
"""
        raise NodeException(
            code=WorkflowErrorCode.INVALID_CODE,
            message=error_message,
        )

    logs = log_buffer.getvalue()
    result = exec_globals["__arg__out"]

    result = cast_to_output_type(result, output_type)

    return logs, result



"""Utilities for reading and managing workflow metadata.json files."""

import json
import os
from uuid import UUID
from typing import Dict, Optional, Type

from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.utils.files import virtual_open


def find_workflow_root_with_metadata(module_path: str) -> Optional[str]:
    """
    Find the workflow root module by searching for metadata.json up the module hierarchy.

    Args:
        module_path: The module path to search from (e.g., "workflows.my_workflow.nodes.my_node")

    Returns:
        The workflow root module path if found, None otherwise
    """
    parts = module_path.split(".")
    for i in range(len(parts), 0, -1):
        potential_root = ".".join(parts[:i])
        file_path = os.path.join(potential_root.replace(".", os.path.sep), "metadata.json")
        try:
            f = virtual_open(file_path)
            if f is not None:
                f.close()
                return potential_root
        except (FileNotFoundError, OSError):
            continue
    return None


def load_edges_to_id_mapping(module_path: str) -> Dict[str, str]:
    """
    Load edge path to ID mapping from metadata.json for a given module.

    This function searches up the module hierarchy for metadata.json and extracts
    the edges_to_id_mapping.

    Args:
        module_path: The module path to search from (e.g., "workflows.my_workflow")

    Returns:
        Dictionary mapping edge keys to their UUID strings
    """
    try:
        root = find_workflow_root_with_metadata(module_path)
        if not root:
            return {}
        file_path = os.path.join(root.replace(".", os.path.sep), "metadata.json")
        with virtual_open(file_path) as f:
            data = json.load(f)
            edges_map = data.get("edges_to_id_mapping")
            return edges_map if isinstance(edges_map, dict) else {}
    except Exception:
        return {}


def get_trigger_edge_id(trigger_cls: Type, target_node: Type[BaseNode], module_path: str) -> Optional[str]:
    """
    Get the stable edge ID for a trigger edge from metadata.json.

    Args:
        trigger_cls: The trigger class
        target_node: The target node class
        module_path: The workflow module path

    Returns:
        The stable edge ID if found in metadata.json, None otherwise
    """
    edges_mapping = load_edges_to_id_mapping(module_path)
    trigger_path = f"{trigger_cls.__module__}.{trigger_cls.__name__}"
    target_path = f"{target_node.__module__}.{target_node.__name__}.Trigger"
    edge_key = f"{trigger_path}|{target_path}"
    return edges_mapping.get(edge_key)


def get_entrypoint_edge_id(target_node: Type[BaseNode], module_path: str) -> Optional[str]:
    """
    Get the stable edge ID for an entrypoint edge from metadata.json.

    Args:
        target_node: The target node class
        module_path: The workflow module path

    Returns:
        The stable edge ID if found in metadata.json, None otherwise
    """
    edges_mapping = load_edges_to_id_mapping(module_path)
    manual_path = "vellum.workflows.triggers.manual.Manual"
    target_path = f"{target_node.__module__}.{target_node.__name__}.Trigger"
    edge_key = f"{manual_path}|{target_path}"
    return edges_mapping.get(edge_key)


def get_regular_edge_id(
    source_node_cls: Type[BaseNode], source_handle_id: UUID, target_node: Type[BaseNode], module_path: str
) -> Optional[str]:
    """
    Get the stable edge ID for a regular node-to-node edge from metadata.json.

    Args:
        source_node_cls: The source node class
        source_handle_id: The source port/handle ID
        target_node: The target node class
        module_path: The workflow module path

    Returns:
        The stable edge ID if found in metadata.json, None otherwise
    """
    edges_mapping = load_edges_to_id_mapping(module_path)
    source_path = f"{source_node_cls.__module__}.{source_node_cls.__name__}.Ports.{str(source_handle_id)}"
    target_path = f"{target_node.__module__}.{target_node.__name__}.Trigger"
    edge_key = f"{source_path}|{target_path}"
    return edges_mapping.get(edge_key)


def load_dataset_row_index_to_id_mapping(module_path: str) -> Dict[int, str]:
    """
    Load dataset row index to ID mapping from metadata.json for a given module.

    This function searches up the module hierarchy for metadata.json and extracts
    the dataset_row_index_to_id_mapping.

    Args:
        module_path: The module path to search from (e.g., "workflows.my_workflow")

    Returns:
        Dictionary mapping dataset row indices (as integers) to their ID strings
    """
    try:
        root = find_workflow_root_with_metadata(module_path)
        if not root:
            return {}
        file_path = os.path.join(root.replace(".", os.path.sep), "metadata.json")
        with virtual_open(file_path) as f:
            data = json.load(f)
            mapping = data.get("dataset_row_index_to_id_mapping")
            if isinstance(mapping, dict):
                return {int(k): v for k, v in mapping.items()}
            return {}
    except Exception:
        return {}



import json
import os
from unittest.mock import patch

from click.testing import CliRunner
from pydash import snake_case

from vellum.workflows.utils.zip import zip_file_map
from vellum_cli import main as cli_main


class MockTemplate:
    def __init__(self, id, label):
        self.id = id
        self.label = label


def test_init_command(vellum_client, mock_module):
    # GIVEN a module on the user's filesystem
    temp_dir = mock_module.temp_dir
    mock_module.set_pyproject_toml({"workflows": []})
    # GIVEN the vellum client returns a list of template workflows
    fake_templates = [
        MockTemplate(id="template-1", label="Example Workflow"),
        MockTemplate(id="template-2", label="Another Workflow"),
    ]
    vellum_client.workflow_sandboxes.list_workflow_sandbox_examples.return_value.results = fake_templates

    # AND the workflow pull API call returns a zip file
    vellum_client.workflows.pull.return_value = iter([zip_file_map({"workflow.py": "print('hello')"})])

    # WHEN the user runs the `init` command and selects the first template
    runner = CliRunner()
    result = runner.invoke(cli_main, ["workflows", "init"], input="1\n")

    # THEN the command returns successfully
    assert result.exit_code == 0

    # AND `vellum_client.workflows.pull` is called with the selected template ID
    vellum_client.workflows.pull.assert_called_once_with(
        "template-1",
        request_options={"additional_query_parameters": {"include_sandbox": True}},
    )

    # AND the `workflow.py` file should be created in the correct module directory
    workflow_py = os.path.join(temp_dir, "example_workflow", "workflow.py")
    assert os.path.exists(workflow_py)
    with open(workflow_py) as f:
        assert f.read() == "print('hello')"

    # AND the vellum.lock.json file should be created
    vellum_lock_json = os.path.join(temp_dir, "vellum.lock.json")
    assert os.path.exists(vellum_lock_json)
    with open(vellum_lock_json) as f:
        lock_data = json.load(f)
        assert lock_data["workflows"] == [
            {
                "module": "example_workflow",
                "workflow_sandbox_id": "template-1",
                "ignore": None,
                "deployments": [],
                "container_image_name": None,
                "container_image_tag": None,
                "workspace": "default",
                "target_directory": None,
            }
        ]


def test_init_command__invalid_template_id(vellum_client, mock_module):
    # GIVEN a module on the user's filesystem
    temp_dir = mock_module.temp_dir
    mock_module.set_pyproject_toml({"workflows": []})
    # GIVEN the vellum client returns a list of template workflows
    fake_templates = [
        MockTemplate(id="template-1", label="Example Workflow"),
        MockTemplate(id="template-2", label="Another Workflow"),
    ]
    vellum_client.workflow_sandboxes.list_workflow_sandbox_examples.return_value.results = fake_templates

    # WHEN the user runs the `init` command, enters invalid input and then cancels
    runner = CliRunner()
    # Mock click.prompt to raise a KeyboardInterrupt (simulating Ctrl+C)
    with patch("click.prompt", side_effect=KeyboardInterrupt):
        runner = CliRunner()
        result = runner.invoke(cli_main, ["workflows", "init"])

    # THEN the command is aborted
    assert result.exit_code != 0
    assert "Aborted!" in result.output  # Click shows this message on Ctrl+C

    # AND `vellum_client.workflows.pull` is not called
    vellum_client.workflows.pull.assert_not_called()

    # AND no workflow files are created
    workflow_py = os.path.join(temp_dir, "example_workflow", "workflow.py")
    assert not os.path.exists(workflow_py)

    # AND the lock file remains empty
    vellum_lock_json = os.path.join(temp_dir, "vellum.lock.json")
    if os.path.exists(vellum_lock_json):
        with open(vellum_lock_json) as f:
            lock_data = json.load(f)
            assert lock_data["workflows"] == []


def test_init_command__no_templates(vellum_client, mock_module):
    # GIVEN a module on the user's filesystem
    temp_dir = mock_module.temp_dir
    mock_module.set_pyproject_toml({"workflows": []})
    # GIVEN the vellum client returns no template workflows
    vellum_client.workflow_sandboxes.list_workflow_sandbox_examples.return_value.results = []

    # WHEN the user runs the `init` command
    runner = CliRunner()
    result = runner.invoke(cli_main, ["workflows", "init"])

    # THEN the command gracefully exits
    assert result.exit_code == 0
    assert "No templates available" in result.output

    # AND `vellum_client.workflows.pull` is not called
    vellum_client.workflows.pull.assert_not_called()

    # AND no workflow files are created
    workflow_py = os.path.join(temp_dir, "example_workflow", "workflow.py")
    assert not os.path.exists(workflow_py)

    # AND the lock file remains empty
    vellum_lock_json = os.path.join(temp_dir, "vellum.lock.json")
    if os.path.exists(vellum_lock_json):
        with open(vellum_lock_json) as f:
            lock_data = json.load(f)
            assert lock_data["workflows"] == []


def test_init_command_target_directory_exists(vellum_client, mock_module):
    """
    GIVEN a target directory already exists
    WHEN the user tries to run the `init` command
    THEN the command should fail and exit without modifying existing files
    """
    temp_dir = mock_module.temp_dir
    existing_workflow_dir = os.path.join(temp_dir, "example_workflow")

    # Create the target directory to simulate it already existing
    os.makedirs(existing_workflow_dir, exist_ok=True)

    # Ensure directory exists before command execution
    assert os.path.exists(existing_workflow_dir)

    # GIVEN the vellum client returns a list of template workflows
    fake_templates = [
        MockTemplate(id="template-1", label="Example Workflow"),
    ]
    vellum_client.workflow_sandboxes.list_workflow_sandbox_examples.return_value.results = fake_templates

    # AND the workflow pull API call returns a zip file
    vellum_client.workflows.pull.return_value = iter([zip_file_map({"workflow.py": "print('hello')"})])

    # WHEN the user runs the `init` command and selects the template
    runner = CliRunner()
    result = runner.invoke(cli_main, ["workflows", "init"], input="1\n")

    # THEN the command should detect the existing directory and abort
    assert result.exit_code == 0
    assert f"{existing_workflow_dir} already exists." in result.output

    # Ensure the directory still exists (wasn't deleted or modified)
    assert os.path.exists(existing_workflow_dir)

    # AND `vellum_client.workflows.pull` is not called
    vellum_client.workflows.pull.assert_not_called()

    # AND no workflow files are created
    workflow_py = os.path.join(temp_dir, "example_workflow", "workflow.py")
    assert not os.path.exists(workflow_py)

    # AND the lock file remains empty
    vellum_lock_json = os.path.join(temp_dir, "vellum.lock.json")
    if os.path.exists(vellum_lock_json):
        with open(vellum_lock_json) as f:
            lock_data = json.load(f)
            assert lock_data["workflows"] == []


def test_init_command_with_template_name(vellum_client, mock_module):
    # GIVEN a module on the user's filesystem
    temp_dir = mock_module.temp_dir
    mock_module.set_pyproject_toml({"workflows": []})

    # GIVEN the vellum client returns a list of template workflows
    fake_templates = [
        MockTemplate(id="template-1", label="Example Workflow"),
        MockTemplate(id="template-2", label="Another Workflow"),
    ]
    vellum_client.workflow_sandboxes.list_workflow_sandbox_examples.return_value.results = fake_templates

    # AND the workflow pull API call returns a zip file
    vellum_client.workflows.pull.return_value = iter([zip_file_map({"workflow.py": "print('hello')"})])

    # WHEN the user runs the `init` command with a specific template name
    template_name = snake_case("Another Workflow")
    runner = CliRunner()
    result = runner.invoke(cli_main, ["workflows", "init", template_name])

    # THEN the command returns successfully
    assert result.exit_code == 0

    # AND `vellum_client.workflows.pull` is called with the correct template ID
    vellum_client.workflows.pull.assert_called_once_with(
        "template-2",  # ID of "Another Workflow"
        request_options={"additional_query_parameters": {"include_sandbox": True}},
    )

    # AND the workflow files should be created in the correct module directory
    workflow_py = os.path.join(temp_dir, "another_workflow", "workflow.py")

    assert os.path.exists(workflow_py)

    with open(workflow_py) as f:
        assert f.read() == "print('hello')"

    # AND the vellum.lock.json file should be created with the correct data
    vellum_lock_json = os.path.join(temp_dir, "vellum.lock.json")
    assert os.path.exists(vellum_lock_json)

    with open(vellum_lock_json) as f:
        lock_data = json.load(f)
        assert lock_data["workflows"] == [
            {
                "module": "another_workflow",
                "workflow_sandbox_id": "template-2",
                "ignore": None,
                "deployments": [],
                "container_image_name": None,
                "container_image_tag": None,
                "workspace": "default",
                "target_directory": None,
            }
        ]


def test_init_command_with_nonexistent_template_name(vellum_client, mock_module):
    # GIVEN a module on the user's filesystem
    temp_dir = mock_module.temp_dir
    mock_module.set_pyproject_toml({"workflows": []})

    # GIVEN the vellum client returns a list of template workflows
    fake_templates = [
        MockTemplate(id="template-1", label="Example Workflow"),
        MockTemplate(id="template-2", label="Another Workflow"),
    ]
    vellum_client.workflow_sandboxes.list_workflow_sandbox_examples.return_value.results = fake_templates

    # WHEN the user runs the `init` command with a non-existent template name
    nonexistent_template = "nonexistent_template"
    runner = CliRunner()
    result = runner.invoke(cli_main, ["workflows", "init", nonexistent_template])

    # THEN the command should indicate the template was not found
    assert result.exit_code == 0
    assert f"Template {nonexistent_template} not found" in result.output

    # AND `vellum_client.workflows.pull` is not called
    vellum_client.workflows.pull.assert_not_called()

    # AND no workflow files are created
    example_workflow_dir = os.path.join(temp_dir, "example_workflow")
    nonexistent_workflow_dir = os.path.join(temp_dir, nonexistent_template)

    assert not os.path.exists(example_workflow_dir)
    assert not os.path.exists(nonexistent_workflow_dir)

    # AND the lock file remains empty
    vellum_lock_json = os.path.join(temp_dir, "vellum.lock.json")
    if os.path.exists(vellum_lock_json):
        with open(vellum_lock_json) as f:
            lock_data = json.load(f)
            assert lock_data["workflows"] == []


def test_init__with_target_dir(vellum_client, mock_module):
    # GIVEN a module on the user's filesystem
    temp_dir = mock_module.temp_dir
    mock_module.set_pyproject_toml({"workflows": []})

    # GIVEN the vellum client returns a list of template workflows
    fake_templates = [
        MockTemplate(id="template-1", label="Example Workflow"),
    ]
    vellum_client.workflow_sandboxes.list_workflow_sandbox_examples.return_value.results = fake_templates

    # AND the workflow pull API call returns a zip file
    vellum_client.workflows.pull.return_value = iter([zip_file_map({"workflow.py": "print('hello')"})])

    # AND a target directory
    target_dir = os.path.join(temp_dir, "dir")
    os.makedirs(target_dir, exist_ok=True)

    # WHEN the user runs the init command with target-dir
    runner = CliRunner()
    result = runner.invoke(cli_main, ["workflows", "init", "--target-dir", target_dir], input="1\n")

    # THEN the command returns successfully
    assert result.exit_code == 0

    # AND the `workflow.py` file should be created in the target directory
    module_path = os.path.join(target_dir, "example_workflow")
    workflow_py = os.path.join(module_path, "workflow.py")
    assert os.path.exists(workflow_py)
    with open(workflow_py) as f:
        assert f.read() == "print('hello')"

    # AND the files are not in the default module directory
    default_module_path = os.path.join(temp_dir, "example_workflow", "workflow.py")
    assert not os.path.exists(default_module_path)

    # AND the vellum.lock.json file should be created in the original directory
    vellum_lock_json = os.path.join(temp_dir, "vellum.lock.json")
    assert os.path.exists(vellum_lock_json)
    with open(vellum_lock_json) as f:
        lock_data = json.load(f)
        assert lock_data["workflows"] == [
            {
                "module": "example_workflow",
                "workflow_sandbox_id": "template-1",
                "ignore": None,
                "deployments": [],
                "container_image_name": None,
                "container_image_tag": None,
                "workspace": "default",
                "target_directory": module_path,
            }
        ]


def test_init__with_nested_target_dir(vellum_client, mock_module):
    # GIVEN a module on the user's filesystem
    temp_dir = mock_module.temp_dir
    mock_module.set_pyproject_toml({"workflows": []})

    # GIVEN the vellum client returns a list of template workflows
    fake_templates = [
        MockTemplate(id="template-1", label="Example Workflow"),
    ]
    vellum_client.workflow_sandboxes.list_workflow_sandbox_examples.return_value.results = fake_templates

    # AND the workflow pull API call returns a zip file
    vellum_client.workflows.pull.return_value = iter([zip_file_map({"workflow.py": "print('hello')"})])

    # AND a nested target directory that doesn't exist yet
    nested_target_dir = os.path.join(temp_dir, "dir-1", "dir-2")

    # WHEN the user runs the init command with nested target-dir
    runner = CliRunner()
    result = runner.invoke(cli_main, ["workflows", "init", "--target-dir", nested_target_dir], input="1\n")

    # THEN the command returns successfully
    assert result.exit_code == 0

    # AND the nested directory with module subdirectory should be created
    module_path = os.path.join(nested_target_dir, "example_workflow")
    assert os.path.exists(module_path)

    # AND the workflow.py file is written to the nested target directory
    workflow_py = os.path.join(module_path, "workflow.py")
    assert os.path.exists(workflow_py)
    with open(workflow_py) as f:
        assert f.read() == "print('hello')"

    # AND the files are not in the default module directory
    default_module_path = os.path.join(temp_dir, "example_workflow", "workflow.py")
    assert not os.path.exists(default_module_path)

    # AND the vellum.lock.json file is still updated
    vellum_lock_json = os.path.join(temp_dir, "vellum.lock.json")
    assert os.path.exists(vellum_lock_json)
    with open(vellum_lock_json) as f:
        lock_data = json.load(f)
        assert lock_data["workflows"] == [
            {
                "module": "example_workflow",
                "workflow_sandbox_id": "template-1",
                "ignore": None,
                "deployments": [],
                "container_image_name": None,
                "container_image_tag": None,
                "workspace": "default",
                "target_directory": module_path,
            }
        ]


def test_init__with_template_name_and_target_dir(vellum_client, mock_module):
    # GIVEN a module on the user's filesystem
    temp_dir = mock_module.temp_dir
    mock_module.set_pyproject_toml({"workflows": []})

    # GIVEN the vellum client returns a list of template workflows
    fake_templates = [
        MockTemplate(id="template-1", label="Example Workflow"),
        MockTemplate(id="template-2", label="Another Workflow"),
    ]
    vellum_client.workflow_sandboxes.list_workflow_sandbox_examples.return_value.results = fake_templates

    # AND the workflow pull API call returns a zip file
    vellum_client.workflows.pull.return_value = iter([zip_file_map({"workflow.py": "print('hello')"})])

    # AND a target directory
    target_dir = os.path.join(temp_dir, "dir")
    os.makedirs(target_dir, exist_ok=True)

    # WHEN the user runs the init command with a specific template name and target-dir
    template_name = snake_case("Another Workflow")
    runner = CliRunner()
    result = runner.invoke(cli_main, ["workflows", "init", template_name, "--target-dir", target_dir])

    # THEN the command returns successfully
    assert result.exit_code == 0

    # AND `vellum_client.workflows.pull` is called with the correct template ID
    vellum_client.workflows.pull.assert_called_once_with(
        "template-2",  # ID of "Another Workflow"
        request_options={"additional_query_parameters": {"include_sandbox": True}},
    )

    # AND the workflow files should be created in the target directory with the correct module subdirectory
    module_path = os.path.join(target_dir, "another_workflow")
    workflow_py = os.path.join(module_path, "workflow.py")
    assert os.path.exists(workflow_py)
    with open(workflow_py) as f:
        assert f.read() == "print('hello')"

    # AND the files are not in the default module directory
    default_module_path = os.path.join(temp_dir, "another_workflow", "workflow.py")
    assert not os.path.exists(default_module_path)

    # AND the vellum.lock.json file should be created with the correct data
    vellum_lock_json = os.path.join(temp_dir, "vellum.lock.json")
    assert os.path.exists(vellum_lock_json)

    with open(vellum_lock_json) as f:
        lock_data = json.load(f)
        assert lock_data["workflows"] == [
            {
                "module": "another_workflow",
                "workflow_sandbox_id": "template-2",
                "ignore": None,
                "deployments": [],
                "container_image_name": None,
                "container_image_tag": None,
                "workspace": "default",
                "target_directory": module_path,
            }
        ]



from __future__ import annotations

import pytest
from typing import Any, ClassVar, Generic, List, TypeVar, Union

from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.nodes.core.try_node.node import TryNode
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.references.output import OutputReference
from vellum.workflows.types.utils import get_class_attr_names, infer_types


class ExampleClass:
    alpha: str
    beta = 3
    gamma: Union[str, int]
    epsilon = OutputReference(
        name="epsilon",
        types=(List[str],),
        instance=None,
        outputs_class=BaseOutputs,
    )
    zeta: ClassVar[str]
    eta: List[str]
    kappa: Any
    mu: list[str]


class ExamplePEP604Class:
    pep604_union: str | int
    pep604_optional: str | None


T = TypeVar("T")


class ExampleGenericClass(Generic[T]):
    delta: T


class ExampleInheritedClass(ExampleClass):
    theta: int


@TryNode.wrap()
class ExampleNode(BaseNode):
    class Outputs(BaseNode.Outputs):
        iota: str


@pytest.mark.parametrize(
    "cls, attr_name, expected_type",
    [
        (ExampleClass, "alpha", (str,)),
        (ExampleClass, "beta", (int,)),
        (ExampleClass, "gamma", (str, int)),
        (ExampleGenericClass, "delta", ()),
        (ExampleGenericClass[str], "delta", (str,)),
        (ExampleClass, "epsilon", (List[str],)),
        (ExampleClass, "zeta", (str,)),
        (ExampleClass, "eta", (List[str],)),
        (ExampleInheritedClass, "theta", (int,)),
        (ExampleInheritedClass, "alpha", (str,)),
        (ExampleInheritedClass, "beta", (int,)),
        (ExampleNode.Outputs, "iota", (str,)),
        (ExampleClass, "kappa", (Any,)),
        (ExampleClass, "mu", (list[str],)),
        (ExamplePEP604Class, "pep604_union", (str, int)),
        (ExamplePEP604Class, "pep604_optional", (str, type(None))),
    ],
    ids=[
        "str",
        "int",
        "str_or_int",
        "generic_blank",
        "generic_str",
        "descriptor",
        "class_var",
        "list_str",
        "inherited_int",
        "inherited_parent_annotation",
        "inherited_parent_class_var",
        "try_node_output",
        "any",
        "list_str_generic",
        "pep604_union",
        "pep604_optional",
    ],
)
def test_infer_types(cls, attr_name, expected_type):
    assert infer_types(cls, attr_name) == expected_type


@pytest.mark.parametrize(
    "cls, expected_attr_names",
    [
        (ExampleClass, ["beta", "epsilon", "alpha", "gamma", "zeta", "eta", "kappa", "mu"]),
        (ExampleGenericClass, ["delta"]),
        (ExampleInheritedClass, ["theta", "beta", "epsilon", "alpha", "gamma", "zeta", "eta", "kappa", "mu"]),
    ],
)
def test_class_attr_names(cls, expected_attr_names):
    assert get_class_attr_names(cls) == expected_attr_names



from uuid import UUID, uuid4

from vellum.workflows import BaseWorkflow
from vellum.workflows.context import execution_context, get_execution_context
from vellum.workflows.events.types import NodeParentContext, WorkflowParentContext
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes import BaseNode
from vellum.workflows.references import VellumSecretReference
from vellum.workflows.state import BaseState


class MockInputs(BaseInputs):
    foo: str


class MockNode(BaseNode):
    node_foo = MockInputs.foo
    node_secret = VellumSecretReference("secret")

    class Outputs(BaseNode.Outputs):
        example: str


class MockWorkflow(BaseWorkflow[MockInputs, BaseState]):
    graph = MockNode


def test_context_trace_and_parent():
    trace_id = uuid4()
    parent_context = NodeParentContext(
        node_definition=MockNode,
        span_id=UUID("123e4567-e89b-12d3-a456-426614174000"),
        parent=WorkflowParentContext(
            workflow_definition=MockWorkflow,
            span_id=UUID("123e4567-e89b-12d3-a456-426614174000"),
        ),
    )
    second_parent_context = WorkflowParentContext(
        workflow_definition=MockWorkflow, span_id=uuid4(), parent=parent_context
    )
    # When using execution context , if we set trace id within
    with execution_context(parent_context=parent_context, trace_id=trace_id):
        test = get_execution_context()
        assert test.trace_id == trace_id
        assert test.parent_context == parent_context
        with execution_context(parent_context=second_parent_context):
            test1 = get_execution_context()
            assert test1.trace_id == trace_id
            assert test1.parent_context == second_parent_context
            # then we can assume trace id will not change
            with execution_context(trace_id=uuid4()):
                test3 = get_execution_context()
                assert test3.trace_id == trace_id
            with execution_context(parent_context=parent_context, trace_id=uuid4()):
                test3 = get_execution_context()
                assert test3.trace_id == trace_id
    # and if we have a new context, the trace will differ
    with execution_context(parent_context=parent_context, trace_id=uuid4()):
        test = get_execution_context()
        assert test.trace_id != trace_id



from typing import Any, Dict, Generic, Optional, Union

from requests import Request, RequestException, Session
from requests.exceptions import JSONDecodeError

from vellum.client import ApiError
from vellum.client.core.request_options import RequestOptions
from vellum.client.types.method_enum import MethodEnum
from vellum.client.types.vellum_secret import VellumSecret as ClientVellumSecret
from vellum.workflows.constants import APIRequestMethod
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.outputs import BaseOutputs
from vellum.workflows.types.core import Json, MergeBehavior, VellumSecret
from vellum.workflows.types.generics import StateType
from vellum.workflows.utils.hmac import sign_request_with_env_secret


class BaseAPINode(BaseNode, Generic[StateType]):
    """
    Used to execute an API call.

    url: str - The URL to send the request to.
    method: Union[APIRequestMethod, str] - The HTTP method to use for the request.
    data: Optional[str] - The data to send in the request body.
    json: Optional["JsonObject"] - The JSON data to send in the request body.
    headers: Optional[Dict[str, Union[str, VellumSecret]]] - The headers to send in the request.
    timeout: Optional[int] - The timeout in seconds for the API request.
    """

    class Trigger(BaseNode.Trigger):
        merge_behavior = MergeBehavior.AWAIT_ANY

    url: str = ""
    method: Optional[Union[APIRequestMethod, MethodEnum]] = APIRequestMethod.GET
    data: Optional[str] = None
    json: Optional[Json] = None
    headers: Optional[Dict[str, Union[str, VellumSecret]]] = None
    timeout: Optional[int] = None

    class Outputs(BaseOutputs):
        json: Optional[Json]
        headers: Dict[str, str]
        status_code: int
        text: str

    def _normalize_http_method(self, method: Union[APIRequestMethod, MethodEnum]) -> str:
        if isinstance(method, APIRequestMethod):
            method_str = method.value
        elif isinstance(method, str):
            method_str = method.upper()
            valid_methods = {m.value for m in APIRequestMethod}
            if method_str not in valid_methods:
                raise NodeException(f"Invalid HTTP method '{method}'", code=WorkflowErrorCode.INVALID_INPUTS)
        else:
            raise NodeException(
                f"Method must be either APIRequestMethod enum or string, got {type(method)}",
                code=WorkflowErrorCode.INVALID_INPUTS,
            )
        return method_str

    def _validate(self) -> None:
        if not self.url or not isinstance(self.url, str) or not self.url.strip():
            raise NodeException("URL is required and must be a non-empty string", code=WorkflowErrorCode.INVALID_INPUTS)

    def run(self) -> Outputs:
        return self._run(
            method=self.method, url=self.url, data=self.data, json=self.json, headers=self.headers, timeout=self.timeout
        )

    def _run(
        self,
        url: str,
        method: Optional[Union[APIRequestMethod, MethodEnum]] = APIRequestMethod.GET,
        data: Optional[Union[str, Any]] = None,
        json: Any = None,
        headers: Any = None,
        bearer_token: Optional[VellumSecret] = None,
        timeout: Optional[int] = None,
    ) -> Outputs:
        self._validate()

        normalized_method = self._normalize_http_method(method) if method is not None else APIRequestMethod.GET.value

        vellum_instance = False
        for header in headers or {}:
            if isinstance(headers[header], VellumSecret):
                vellum_instance = True
        if vellum_instance or bearer_token:
            return self._vellum_execute_api(bearer_token, json, headers, normalized_method, url, timeout)
        else:
            return self._local_execute_api(data, headers, json, normalized_method, url, timeout)

    def _local_execute_api(self, data, headers, json, method, url, timeout):
        headers = headers or {}
        if "User-Agent" not in headers:
            client_headers = self._context.vellum_client._client_wrapper.get_headers()
            headers["User-Agent"] = client_headers.get("User-Agent")
        try:
            if data is not None:
                prepped = Request(method=method, url=url, data=data, headers=headers).prepare()
            elif json is not None:
                prepped = Request(method=method, url=url, json=json, headers=headers).prepare()
            else:
                prepped = Request(method=method, url=url, headers=headers).prepare()
        except Exception as e:
            raise NodeException(f"Failed to prepare HTTP request: {e}", code=WorkflowErrorCode.PROVIDER_ERROR)

        sign_request_with_env_secret(prepped)

        try:
            with Session() as session:
                response = session.send(prepped, timeout=timeout)
        except RequestException as e:
            raise NodeException(f"HTTP request failed: {e}", code=WorkflowErrorCode.PROVIDER_ERROR)
        try:
            json_response = response.json()
        except JSONDecodeError:
            json_response = None
        return self.Outputs(
            json=json_response,
            headers={header: value for header, value in response.headers.items()},
            status_code=response.status_code,
            text=response.text,
        )

    def _vellum_execute_api(self, bearer_token, data, headers, method, url, timeout):
        client_vellum_secret = ClientVellumSecret(name=bearer_token.name) if bearer_token else None

        headers = headers or {}
        if "User-Agent" not in headers:
            client_headers = self._context.vellum_client._client_wrapper.get_headers()
            headers["User-Agent"] = client_headers.get("User-Agent")

        # Create request_options if timeout is specified
        request_options = None
        if timeout is not None:
            request_options = RequestOptions(timeout_in_seconds=timeout)

        try:
            vellum_response = self._context.vellum_client.execute_api(
                url=url,
                method=method,
                body=data,
                headers=headers,
                bearer_token=client_vellum_secret,
                request_options=request_options,
            )
        except ApiError as e:
            raise NodeException(f"Failed to prepare HTTP request: {e}", code=WorkflowErrorCode.NODE_EXECUTION)

        return self.Outputs(
            json=vellum_response.json_,
            headers={header: value for header, value in vellum_response.headers.items()},
            status_code=vellum_response.status_code,
            text=vellum_response.text,
        )



# This file was auto-generated by Fern from our API Definition.

from .pdf_search_result_meta_source import PdfSearchResultMetaSource

ComponentsSchemasPdfSearchResultMetaSource = PdfSearchResultMetaSource



# This file was auto-generated by Fern from our API Definition.

import typing

from .prompt_request_input import PromptRequestInput
from .workflow_input import WorkflowInput

DeprecatedPromptRequestInput = typing.Union[WorkflowInput, PromptRequestInput]



"""Tests for IntegrationTrigger base class."""

from typing import Any

from vellum.workflows.constants import VellumIntegrationProviderType
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.references.trigger import TriggerAttributeReference
from vellum.workflows.state.base import BaseState
from vellum.workflows.triggers.integration import IntegrationTrigger


def test_integration_trigger__can_be_instantiated_as_base():
    """IntegrationTrigger can be instantiated as a base class."""
    # WHEN we instantiate IntegrationTrigger directly
    trigger = IntegrationTrigger(test="data")

    # THEN it creates an instance with event data stored
    assert trigger._event_data == {"test": "data"}


def test_integration_trigger__can_be_instantiated():
    """IntegrationTrigger can be instantiated for testing."""

    # GIVEN IntegrationTrigger with concrete implementation
    class TestTrigger(IntegrationTrigger):
        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "test"
            slug = "test"

    # THEN it can be instantiated (even though base is ABC, concrete subclasses work)
    assert TestTrigger is not None


def test_integration_trigger__can_be_subclassed():
    """IntegrationTrigger can be subclassed to create concrete triggers."""

    # GIVEN a concrete implementation of IntegrationTrigger
    class TestTrigger(IntegrationTrigger):
        data: str

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "test"
            slug = "test"

        def __init__(self, **kwargs: Any):
            super().__init__(**kwargs)
            self.data = kwargs.get("data", "")

    # WHEN we create a trigger instance
    result = TestTrigger(data="test")

    # THEN it returns the expected trigger instance with populated attributes
    assert result.data == "test"


def test_integration_trigger__attribute_reference():
    """Trigger annotations expose TriggerAttributeReference descriptors."""

    class TestTrigger(IntegrationTrigger):
        value: str

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "test"
            slug = "test"

        def __init__(self, **kwargs: Any):
            super().__init__(**kwargs)
            self.value = kwargs.get("value", "")

    reference = TestTrigger.value
    assert isinstance(reference, TriggerAttributeReference)
    assert reference.name == "value"
    assert TestTrigger.value == reference
    assert reference == TestTrigger.attribute_references()["value"]

    state = BaseState()
    trigger = TestTrigger(value="data")
    trigger.bind_to_state(state)
    assert reference.resolve(state) == "data"


def test_integration_trigger__graph_syntax():
    """IntegrationTrigger can be used in graph syntax."""

    # GIVEN a concrete trigger and a node
    class TestTrigger(IntegrationTrigger):
        value: str

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "test"
            slug = "test"

        def __init__(self, **kwargs: Any):
            super().__init__(**kwargs)
            self.value = kwargs.get("value", "")

    class TestNode(BaseNode):
        pass

    # WHEN we use trigger >> node syntax
    graph = TestTrigger >> TestNode

    # THEN a graph is created
    assert graph is not None
    assert len(list(graph.trigger_edges)) == 1
    assert list(graph.trigger_edges)[0].trigger_class == TestTrigger
    assert list(graph.trigger_edges)[0].to_node == TestNode


def test_integration_trigger__multiple_entrypoints():
    """IntegrationTrigger works with multiple entry points."""

    # GIVEN a trigger and multiple nodes
    class TestTrigger(IntegrationTrigger):
        msg: str

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "test"
            slug = "test"

        def __init__(self, **kwargs: Any):
            super().__init__(**kwargs)
            self.msg = kwargs.get("msg", "")

    class NodeA(BaseNode):
        pass

    class NodeB(BaseNode):
        pass

    # WHEN we use trigger >> {nodes} syntax
    graph = TestTrigger >> {NodeA, NodeB}

    # THEN both nodes are entrypoints
    trigger_edges = list(graph.trigger_edges)
    assert len(trigger_edges) == 2
    target_nodes = {edge.to_node for edge in trigger_edges}
    assert target_nodes == {NodeA, NodeB}


def test_integration_trigger__config_attribute():
    """IntegrationTrigger requires Config class with provider, integration_name, and slug."""

    # GIVEN IntegrationTrigger
    # THEN it has a Config class with type annotations for required fields
    assert hasattr(IntegrationTrigger, "Config")
    assert hasattr(IntegrationTrigger.Config, "__annotations__")
    assert "provider" in IntegrationTrigger.Config.__annotations__
    assert "integration_name" in IntegrationTrigger.Config.__annotations__
    assert "slug" in IntegrationTrigger.Config.__annotations__



from uuid import uuid4

from deepdiff import DeepDiff

from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum_ee.workflows.display.base import WorkflowInputsDisplay
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.types import NodeOutputDisplay


class Inputs(BaseInputs):
    input: str


def test_serialize_node__annotated_output(serialize_node):
    class AnnotatedOutputGenericNode(BaseNode):

        class Outputs(BaseNode.Outputs):
            output: int

    serialized_node = serialize_node(AnnotatedOutputGenericNode)

    assert not DeepDiff(
        {
            "id": "6337349e-5718-4732-b717-b6be650c6664",
            "label": "Annotated Output Generic Node",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "AnnotatedOutputGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_outputs_serialization",
                ],
            },
            "trigger": {"id": "e23d58ad-012b-48a2-80bc-81fe740b5785", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [{"id": "2d86f771-e1b6-4e58-8af1-9b12604e9bb6", "type": "DEFAULT", "name": "default"}],
            "adornments": None,
            "attributes": [],
            "outputs": [
                {
                    "id": "6d08f5f4-cb29-4240-899a-d811e547a30b",
                    "name": "output",
                    "type": "NUMBER",
                    "value": None,
                }
            ],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__workflow_input(serialize_node):
    class WorkflowInputGenericNode(BaseNode):

        class Outputs(BaseNode.Outputs):
            output = Inputs.input

    input_id = uuid4()
    serialized_node = serialize_node(
        node_class=WorkflowInputGenericNode,
        global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=input_id)},
    )

    assert not DeepDiff(
        {
            "id": "b8b45ce8-b897-4347-a8ad-45e0576fcf59",
            "label": "Workflow Input Generic Node",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "WorkflowInputGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_outputs_serialization",
                ],
            },
            "trigger": {"id": "6fb30c9c-b9f0-452f-99f5-58d484b89882", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [{"id": "8997d239-eefb-4006-8961-a8241829b433", "type": "DEFAULT", "name": "default"}],
            "adornments": None,
            "attributes": [],
            "outputs": [
                {
                    "id": "70c7e4c8-2ad3-4093-b69f-a83b4e2a8167",
                    "name": "output",
                    "type": "STRING",
                    "value": {
                        "type": "WORKFLOW_INPUT",
                        "input_variable_id": str(input_id),
                    },
                }
            ],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__node_output_reference(serialize_node):
    class NodeWithOutput(BaseNode):

        class Outputs(BaseNode.Outputs):
            output = Inputs.input

    class NodeWithOutputDisplay(BaseNodeDisplay[NodeWithOutput]):
        pass

    class GenericNodeReferencingOutput(BaseNode):

        class Outputs(BaseNode.Outputs):
            output = NodeWithOutput.Outputs.output

    workflow_input_id = uuid4()
    node_output_id = uuid4()
    serialized_node = serialize_node(
        node_class=GenericNodeReferencingOutput,
        global_workflow_input_displays={Inputs.input: WorkflowInputsDisplay(id=workflow_input_id)},
        global_node_displays={NodeWithOutput: NodeWithOutputDisplay()},
        global_node_output_displays={
            NodeWithOutput.Outputs.output: NodeOutputDisplay(id=node_output_id, name="output")
        },
    )

    assert not DeepDiff(
        {
            "id": "a3a63eb0-ed79-41c1-9bd2-c4b3d69c3017",
            "label": "Generic Node Referencing Output",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "GenericNodeReferencingOutput",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_outputs_serialization",
                ],
            },
            "trigger": {"id": "b12c8ed6-42e2-4dad-9a39-adb1650a6f85", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [{"id": "3953fe95-7ea9-49ce-a45f-cc55c0f0b9a4", "type": "DEFAULT", "name": "default"}],
            "adornments": None,
            "attributes": [],
            "outputs": [
                {
                    "id": "72f37d68-c927-4f4e-8b25-72e8cb2c7f5c",
                    "name": "output",
                    "type": "STRING",
                    "value": {
                        "type": "NODE_OUTPUT",
                        "node_id": "40fcba1f-9b25-4fed-8f15-a2fd80ff85a1",
                        "node_output_id": str(node_output_id),
                    },
                }
            ],
        },
        serialized_node,
        ignore_order=True,
    )



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class FastEmbedVectorizerBaaiBgeSmallEnV15(UniversalBaseModel):
    """
    FastEmbed vectorizer for BAAI/bge-small-en-v1.5.
    """

    model_name: typing.Literal["BAAI/bge-small-en-v1.5"] = "BAAI/bge-small-en-v1.5"

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from pydantic import Field

from .node_execution_fulfilled_event import NodeExecutionFulfilledEvent
from .node_execution_initiated_event import NodeExecutionInitiatedEvent
from .node_execution_paused_event import NodeExecutionPausedEvent
from .node_execution_rejected_event import NodeExecutionRejectedEvent
from .node_execution_resumed_event import NodeExecutionResumedEvent
from .node_execution_streaming_event import NodeExecutionStreamingEvent
from .workflow_execution_fulfilled_event import WorkflowExecutionFulfilledEvent
from .workflow_execution_initiated_event import WorkflowExecutionInitiatedEvent
from .workflow_execution_paused_event import WorkflowExecutionPausedEvent
from .workflow_execution_rejected_event import WorkflowExecutionRejectedEvent
from .workflow_execution_resumed_event import WorkflowExecutionResumedEvent
from .workflow_execution_snapshotted_event import WorkflowExecutionSnapshottedEvent
from .workflow_execution_streaming_event import WorkflowExecutionStreamingEvent

WorkflowEvent = typing.Annotated[
    typing.Union[
        NodeExecutionInitiatedEvent,
        NodeExecutionStreamingEvent,
        NodeExecutionFulfilledEvent,
        NodeExecutionRejectedEvent,
        NodeExecutionPausedEvent,
        NodeExecutionResumedEvent,
        WorkflowExecutionInitiatedEvent,
        WorkflowExecutionStreamingEvent,
        WorkflowExecutionRejectedEvent,
        WorkflowExecutionFulfilledEvent,
        WorkflowExecutionPausedEvent,
        WorkflowExecutionResumedEvent,
        WorkflowExecutionSnapshottedEvent,
    ],
    Field(discriminator="name")
]



from uuid import UUID
from typing import Any, Dict, List, cast

from vellum.workflows.constants import VellumIntegrationProviderType
from vellum.workflows.graph import Graph
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes import BaseNode, InlineSubworkflowNode
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.ports.port import Port
from vellum.workflows.references.lazy import LazyReference
from vellum.workflows.state import BaseState
from vellum.workflows.triggers.integration import IntegrationTrigger
from vellum.workflows.workflows.base import BaseWorkflow
from vellum_ee.workflows.display.base import EdgeDisplay, WorkflowInputsDisplay
from vellum_ee.workflows.display.editor.types import NodeDisplayData, NodeDisplayPosition
from vellum_ee.workflows.display.nodes import BaseNodeDisplay
from vellum_ee.workflows.display.workflows.base_workflow_display import BaseWorkflowDisplay
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_base_workflow_display__serialize_empty_workflow():
    # GIVEN an empty workflow
    class ExampleWorkflow(BaseWorkflow):
        pass

    display = get_workflow_display(workflow_class=ExampleWorkflow)

    # WHEN serializing the workflow
    exec_config = display.serialize()

    # THEN it should return the expected config
    assert exec_config == {
        "input_variables": [],
        "state_variables": [],
        "output_variables": [],
        "workflow_raw_data": {
            "definition": {
                "module": ["vellum_ee", "workflows", "display", "tests", "test_base_workflow_display"],
                "name": "ExampleWorkflow",
            },
            "display_data": {"viewport": {"x": 0.0, "y": 0.0, "zoom": 1.0}},
            "edges": [],
            "nodes": [
                {
                    "data": {"label": "Entrypoint Node", "source_handle_id": "0af025a4-3b25-457d-a7ae-e3a7ba15c86c"},
                    "base": None,
                    "definition": None,
                    "display_data": {"position": {"x": 0.0, "y": -50.0}},
                    "id": "3c41cdd9-999a-48b8-9088-f6dfa1369bfd",
                    "inputs": [],
                    "type": "ENTRYPOINT",
                }
            ],
            "output_values": [],
        },
    }


def test_vellum_workflow_display__serialize_input_variables_with_capitalized_variable_override():
    # GIVEN a workflow with input variables
    class Inputs(BaseInputs):
        foo: str

    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            output = Inputs.foo

    class ExampleWorkflow(BaseWorkflow[Inputs, BaseState]):
        graph = StartNode

    class ExampleWorkflowDisplay(BaseWorkflowDisplay[ExampleWorkflow]):
        inputs_display = {
            Inputs.foo: WorkflowInputsDisplay(id=UUID("97b63d71-5413-417f-9cf5-49e1b4fd56e4"), name="Foo")
        }

    display = get_workflow_display(
        base_display_class=ExampleWorkflowDisplay,
        workflow_class=ExampleWorkflow,
    )

    # WHEN serializing the workflow
    exec_config = display.serialize()

    # THEN the input variables are what we expect
    input_variables = exec_config["input_variables"]

    assert input_variables == [
        {
            "id": "97b63d71-5413-417f-9cf5-49e1b4fd56e4",
            "key": "Foo",
            "type": "STRING",
            "default": None,
            "required": True,
            "extensions": {"color": None},
        }
    ]


def test_vellum_workflow_display_serialize_valid_handle_ids_for_base_nodes():
    # GIVEN a workflow between two base nodes
    class StartNode(BaseNode):
        pass

    class EndNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            hello = "world"

    class Workflow(BaseWorkflow):
        graph = StartNode >> EndNode

        class Outputs(BaseWorkflow.Outputs):
            final_value = EndNode.Outputs.hello

    # AND a display class for this workflow
    workflow_display = get_workflow_display(workflow_class=Workflow)

    # WHEN we serialize the workflow
    exec_config = workflow_display.serialize()

    # THEN the serialized workflow handle ids are valid
    raw_data = exec_config.get("workflow_raw_data")
    assert isinstance(raw_data, dict)
    nodes = raw_data.get("nodes")
    edges = raw_data.get("edges")

    assert isinstance(nodes, list)
    assert isinstance(edges, list)

    edge_source_handle_ids = {edge.get("source_handle_id") for edge in edges if isinstance(edge, dict)}
    edge_target_handle_ids = {edge.get("target_handle_id") for edge in edges if isinstance(edge, dict)}

    for node in nodes:
        assert isinstance(node, dict)

        if node["type"] in {"ENTRYPOINT", "TERMINAL"}:
            continue

        ports = node.get("ports")
        assert isinstance(ports, list)
        for port in ports:
            assert isinstance(port, dict)
            assert (
                port["id"] in edge_source_handle_ids
            ), f"Port {port['id']} from node {node['label']} not found in edge source handle ids"

        assert isinstance(node["trigger"], dict)
        assert (
            node["trigger"]["id"] in edge_target_handle_ids
        ), f"Trigger {node['trigger']['id']} from node {node['label']} not found in edge target handle ids"


def test_vellum_workflow_display__serialize_with_unused_nodes_and_edges():
    # GIVEN a workflow with active and unused nodes
    class NodeA(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

    class NodeB(BaseNode):
        pass

    class NodeC(BaseNode):
        pass

    # AND A workflow that uses them correctly
    class Workflow(BaseWorkflow):
        graph = NodeA
        unused_graphs = {NodeB >> NodeC}

        class Outputs(BaseWorkflow.Outputs):
            final = NodeA.Outputs.result

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=Workflow)

    # WHEN we serialize the workflow
    exec_config = workflow_display.serialize()

    # THEN the serialized workflow contains the expected nodes and edges
    raw_data = exec_config["workflow_raw_data"]
    assert isinstance(raw_data, dict)

    nodes = raw_data["nodes"]
    edges = raw_data["edges"]

    assert isinstance(nodes, list)
    assert isinstance(edges, list)

    # Find nodes by their definition name
    node_ids: Dict[str, str] = {}

    for node in nodes:
        assert isinstance(node, dict)
        definition = node.get("definition")
        if definition is None:
            continue

        assert isinstance(definition, dict)
        name = definition.get("name")
        if not isinstance(name, str):
            continue

        if name in ["NodeA", "NodeB", "NodeC"]:
            node_id = node.get("id")
            if isinstance(node_id, str):
                node_ids[name] = node_id

    # Verify all nodes are present
    assert "NodeA" in node_ids, "Active node NodeA not found in serialized output"
    assert "NodeB" in node_ids, "Unused node NodeB not found in serialized output"
    assert "NodeC" in node_ids, "Unused node NodeC not found in serialized output"

    # Verify the edge between NodeB and NodeC is present
    edge_found = False
    for edge in edges:
        assert isinstance(edge, dict)
        source_id = edge.get("source_node_id")
        target_id = edge.get("target_node_id")

        if (
            isinstance(source_id, str)
            and isinstance(target_id, str)
            and source_id == node_ids["NodeB"]
            and target_id == node_ids["NodeC"]
        ):
            edge_found = True
            break

    assert edge_found, "Edge between unused nodes NodeB and NodeC not found in serialized output"


def test_vellum_workflow_display__serialize_with_parse_json_expression():
    # GIVEN a workflow that uses the parse_json function
    from vellum.workflows.references.constant import ConstantValueReference

    class JsonNode(BaseNode):

        class Outputs(BaseNode.Outputs):
            json_result = ConstantValueReference('{"key": "value"}').parse_json()

    class Workflow(BaseWorkflow):
        graph = JsonNode

        class Outputs(BaseWorkflow.Outputs):
            final = JsonNode.Outputs.json_result

    # AND a display class for this workflow
    workflow_display = get_workflow_display(workflow_class=Workflow)

    # WHEN we serialize the workflow
    exec_config = workflow_display.serialize()

    # THEN the serialized workflow contains the parse_json expression
    raw_data = exec_config["workflow_raw_data"]
    assert isinstance(raw_data, dict)

    nodes = raw_data["nodes"]
    assert isinstance(nodes, list)

    json_node = None
    for node in nodes:
        assert isinstance(node, dict)
        definition = node.get("definition")
        if node.get("type") == "GENERIC" and isinstance(definition, dict) and definition.get("name") == "JsonNode":
            json_node = node
            break

    assert json_node is not None

    outputs = json_node.get("outputs", [])
    assert isinstance(outputs, list)

    json_result = None
    for output in outputs:
        assert isinstance(output, dict)
        if output.get("name") == "json_result":
            json_result = output
            break

    assert json_result == {
        "id": "e73fd6b1-1109-4a97-8510-c9ba8e6f5dbe",
        "name": "json_result",
        "type": "JSON",
        "value": {
            "type": "UNARY_EXPRESSION",
            "lhs": {
                "type": "CONSTANT_VALUE",
                "value": {
                    "type": "STRING",
                    "value": '{"key": "value"}',
                },
            },
            "operator": "parseJson",
        },
    }


def test_serialize__port_with_lazy_reference():
    # GIVEN a node with a lazy reference in a Port
    class MyNode(BaseNode):

        class Ports(BaseNode.Ports):
            foo = Port.on_if(LazyReference(lambda: MyNode.Outputs.bar))

        class Outputs(BaseNode.Outputs):
            bar: bool

    # AND a workflow that uses the node
    class Workflow(BaseWorkflow):
        graph = MyNode

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=Workflow)
    exec_config = workflow_display.serialize()

    # THEN the lazy reference should be serialized correctly
    raw_data = exec_config["workflow_raw_data"]
    assert isinstance(raw_data, dict)

    nodes = raw_data["nodes"]
    assert isinstance(nodes, list)

    my_node = nodes[1]
    assert isinstance(my_node, dict)
    ports = my_node.get("ports")
    assert isinstance(ports, list)
    assert ports == [
        {
            "id": "02e90580-e6b6-441c-a5d4-7cd083c11fc7",
            "name": "foo",
            "type": "IF",
            "expression": {
                "type": "NODE_OUTPUT",
                "node_id": str(MyNode.__id__),
                "node_output_id": str(MyNode.__output_ids__["bar"]),
            },
        }
    ]


def test_global_propagation_deep_nested_subworkflows():
    # GIVEN the root workflow, a middle workflow, and an inner workflow

    class RootInputs(BaseInputs):
        root_param: str

    class MiddleInputs(BaseInputs):
        middle_param: str

    class InnerInputs(BaseInputs):
        inner_param: str

    class InnerNode(BaseNode):
        class Outputs(BaseOutputs):
            done: bool

        def run(self) -> Outputs:
            return self.Outputs(done=True)

    class InnerWorkflow(BaseWorkflow[InnerInputs, BaseState]):
        graph = InnerNode

    class MiddleInlineSubworkflowNode(InlineSubworkflowNode):
        subworkflow_inputs = {"inner_param": "x"}
        subworkflow = InnerWorkflow

    class MiddleWorkflow(BaseWorkflow[MiddleInputs, BaseState]):
        graph = MiddleInlineSubworkflowNode

    class OuterInlineSubworkflowNode(InlineSubworkflowNode):
        subworkflow_inputs = {"middle_param": "y"}
        subworkflow = MiddleWorkflow

    class RootWorkflow(BaseWorkflow[RootInputs, BaseState]):
        graph = OuterInlineSubworkflowNode

    # WHEN we build the displays
    root_display = get_workflow_display(workflow_class=RootWorkflow)
    middle_display = get_workflow_display(
        workflow_class=MiddleWorkflow, parent_display_context=root_display.display_context
    )
    inner_display = get_workflow_display(
        workflow_class=InnerWorkflow, parent_display_context=middle_display.display_context
    )

    # THEN the deepest display must include root + middle + inner inputs in its GLOBAL view
    inner_global_names = {ref.name for ref in inner_display.display_context.global_workflow_input_displays.keys()}

    assert inner_global_names == {"middle_param", "inner_param", "root_param"}


def test_serialize_workflow_with_edge_display_data():
    """
    Tests that edges with z_index values serialize display_data correctly.
    """

    # GIVEN a workflow with connected nodes
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

    class EndNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            final: str

    class TestWorkflow(BaseWorkflow):
        graph = StartNode >> EndNode

        class Outputs(BaseWorkflow.Outputs):
            final_result = EndNode.Outputs.final

    class TestWorkflowDisplay(BaseWorkflowDisplay[TestWorkflow]):
        edge_displays = {
            (StartNode.Ports.default, EndNode): EdgeDisplay(id=UUID("12345678-1234-5678-1234-567812345678"), z_index=5)
        }

    # WHEN we serialize the workflow with the custom display
    display = get_workflow_display(
        base_display_class=TestWorkflowDisplay,
        workflow_class=TestWorkflow,
    )
    serialized_workflow = display.serialize()

    # THEN the edge should include display_data with z_index
    workflow_raw_data = cast(Dict[str, Any], serialized_workflow["workflow_raw_data"])
    edges = cast(List[Dict[str, Any]], workflow_raw_data["edges"])

    edge_with_display_data = None
    for edge in edges:
        if edge["id"] == "12345678-1234-5678-1234-567812345678":
            edge_with_display_data = edge
            break

    assert edge_with_display_data is not None, "Edge with custom UUID not found"
    assert edge_with_display_data["display_data"] == {"z_index": 5}

    assert edge_with_display_data["type"] == "DEFAULT"
    assert "source_node_id" in edge_with_display_data
    assert "target_node_id" in edge_with_display_data


def test_serialize_workflow_with_node_display_data():
    """
    Tests that nodes with z_index values serialize display_data correctly.
    """

    # GIVEN a workflow with a node that has custom display data
    class TestNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

    class TestWorkflow(BaseWorkflow):
        graph = TestNode

        class Outputs(BaseWorkflow.Outputs):
            final_result = TestNode.Outputs.result

    class TestNodeDisplay(BaseNodeDisplay[TestNode]):
        display_data = NodeDisplayData(position=NodeDisplayPosition(x=100, y=200), z_index=10, width=300, height=150)

    class TestWorkflowDisplay(BaseWorkflowDisplay[TestWorkflow]):
        pass

    # WHEN we serialize the workflow with the custom node display
    display = get_workflow_display(
        base_display_class=TestWorkflowDisplay,
        workflow_class=TestWorkflow,
    )
    serialized_workflow = display.serialize()

    # THEN the node should include display_data with z_index
    workflow_raw_data = cast(Dict[str, Any], serialized_workflow["workflow_raw_data"])
    nodes = cast(List[Dict[str, Any]], workflow_raw_data["nodes"])

    test_node = None
    for node in nodes:
        if node.get("type") == "GENERIC":
            definition = node.get("definition")
            if isinstance(definition, dict) and definition.get("name") == "TestNode":
                test_node = node
                break

    assert test_node is not None, "TestNode not found in serialized nodes"
    assert test_node["display_data"] == {"position": {"x": 100, "y": 200}, "z_index": 10, "width": 300, "height": 150}


def test_serialize_workflow_with_node_icon_and_color():
    """
    Tests that nodes with icon and color serialize correctly in workflow context.
    """

    # GIVEN a workflow with a node that has icon and color
    class TestNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

    class TestWorkflow(BaseWorkflow):
        graph = TestNode

        class Outputs(BaseWorkflow.Outputs):
            final_result = TestNode.Outputs.result

    class TestNodeDisplay(BaseNodeDisplay[TestNode]):
        display_data = NodeDisplayData(position=NodeDisplayPosition(x=100, y=200), icon="vellum:icon:cog", color="navy")

    class TestWorkflowDisplay(BaseWorkflowDisplay[TestWorkflow]):
        pass

    # WHEN we serialize the workflow
    display = get_workflow_display(
        base_display_class=TestWorkflowDisplay,
        workflow_class=TestWorkflow,
    )
    serialized_workflow = display.serialize()

    # THEN the node should include icon and color in display_data
    workflow_raw_data = cast(Dict[str, Any], serialized_workflow["workflow_raw_data"])
    nodes = cast(List[Dict[str, Any]], workflow_raw_data["nodes"])

    test_node = None
    for node in nodes:
        if node.get("type") == "GENERIC":
            definition = node.get("definition")
            if isinstance(definition, dict) and definition.get("name") == "TestNode":
                test_node = node
                break

    assert test_node is not None, "TestNode not found in serialized nodes"
    assert test_node["display_data"]["icon"] == "vellum:icon:cog"
    assert test_node["display_data"]["color"] == "navy"


def test_base_workflow_display__graph_with_trigger_and_regular_node():
    """
    Tests that a workflow with both a trigger edge and a regular node edge serializes correctly.
    """

    # GIVEN a workflow with a trigger and regular nodes
    class SlackMessageTrigger(IntegrationTrigger):
        message: str

        class Config(IntegrationTrigger.Config):
            provider = VellumIntegrationProviderType.COMPOSIO
            integration_name = "SLACK"
            slug = "slack_message"

    class TopNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            result: str

    class BottomNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            final: str

    class TestWorkflow(BaseWorkflow):
        graph = {
            Graph.from_node(TopNode),
            SlackMessageTrigger >> BottomNode,
        }

    # WHEN we serialize the workflow
    workflow_display = get_workflow_display(workflow_class=TestWorkflow)
    exec_config = workflow_display.serialize()

    # THEN the serialized workflow should have the expected structure
    workflow_raw_data = exec_config["workflow_raw_data"]
    assert isinstance(workflow_raw_data, dict)

    nodes = workflow_raw_data["nodes"]
    edges = workflow_raw_data["edges"]
    triggers = exec_config["triggers"]

    assert isinstance(nodes, list)
    assert isinstance(edges, list)
    assert isinstance(triggers, list)

    assert len(triggers) == 1
    trigger = triggers[0]
    assert isinstance(trigger, dict)
    trigger_id = trigger["id"]

    entrypoint_nodes = [n for n in nodes if isinstance(n, dict) and n.get("type") == "ENTRYPOINT"]
    assert len(entrypoint_nodes) == 1
    entrypoint_node = entrypoint_nodes[0]
    assert isinstance(entrypoint_node, dict)
    entrypoint_node_id = entrypoint_node["id"]

    top_node = next(
        (
            node
            for node in nodes
            if isinstance(node, dict)
            and isinstance(node.get("definition"), dict)
            and cast(Dict[str, Any], node.get("definition")).get("name") == "TopNode"
        ),
        None,
    )
    bottom_node = next(
        (
            node
            for node in nodes
            if isinstance(node, dict)
            and isinstance(node.get("definition"), dict)
            and cast(Dict[str, Any], node.get("definition")).get("name") == "BottomNode"
        ),
        None,
    )

    assert top_node is not None, "TopNode not found in serialized nodes"
    assert bottom_node is not None, "BottomNode not found in serialized nodes"
    top_node_id = top_node["id"]
    bottom_node_id = bottom_node["id"]

    entrypoint_edges = [e for e in edges if isinstance(e, dict) and e.get("source_node_id") == entrypoint_node_id]
    assert len(entrypoint_edges) == 1, "Should have exactly one edge from the entrypoint node"

    entrypoint_edge = entrypoint_edges[0]
    assert isinstance(entrypoint_edge, dict)
    assert entrypoint_edge["target_node_id"] == top_node_id, "Entrypoint edge should connect to TopNode"

    trigger_edges = [e for e in edges if isinstance(e, dict) and e.get("source_node_id") == trigger_id]
    assert len(trigger_edges) == 1, "Should have exactly one edge from the trigger"

    trigger_edge = trigger_edges[0]
    assert isinstance(trigger_edge, dict)
    assert trigger_edge["target_node_id"] == bottom_node_id, "Trigger edge should connect to BottomNode"


def test_serialize_subworkflow_output_reference_without_display():
    """
    Tests that a workflow with a subworkflow node followed by a generic node
    that references the subworkflow output can be serialized correctly without a display directory.
    """

    # GIVEN a workflow module with a subworkflow node and a follow-on node
    module_path = "tests.workflows.subworkflow_output_reference_without_display"

    # WHEN we serialize the module
    result = BaseWorkflowDisplay.serialize_module(module_path)

    # THEN it should serialize successfully
    assert hasattr(result, "exec_config")
    assert isinstance(result.exec_config, dict)
    assert "workflow_raw_data" in result.exec_config

    # AND the workflow should have the expected nodes
    workflow_raw_data = result.exec_config["workflow_raw_data"]
    assert isinstance(workflow_raw_data, dict)
    nodes = workflow_raw_data.get("nodes")
    assert isinstance(nodes, list)

    subworkflow_node = next(
        (
            node
            for node in nodes
            if isinstance(node, dict)
            and isinstance(node.get("definition"), dict)
            and node.get("definition", {}).get("name") == "SubworkflowNodeExample"
        ),
        None,
    )
    followon_node = next(
        (
            node
            for node in nodes
            if isinstance(node, dict)
            and isinstance(node.get("definition"), dict)
            and node.get("definition", {}).get("name") == "FollowOnNode"
        ),
        None,
    )

    assert subworkflow_node is not None, "SubworkflowNodeExample not found in serialized nodes"
    assert followon_node is not None, "FollowOnNode not found in serialized nodes"

    # AND the subworkflow node should have output variables
    subworkflow_data = subworkflow_node.get("data", {})
    subworkflow_outputs = subworkflow_data.get("output_variables", [])
    assert len(subworkflow_outputs) > 0, "SubworkflowNodeExample should have output variables"

    result_output = next((out for out in subworkflow_outputs if out.get("key") == "result"), None)
    assert result_output is not None, "SubworkflowNodeExample should have 'result' output"
    expected_output_id = result_output["id"]

    # AND the follow-on node should have an attribute that references the subworkflow output
    followon_attributes = followon_node.get("attributes", [])
    assert len(followon_attributes) > 0, "FollowOnNode should have attributes"

    input_value_attr = next(
        (attr for attr in followon_attributes if isinstance(attr, dict) and attr.get("name") == "input_value"), None
    )

    assert input_value_attr is not None, "input_value attribute not found in FollowOnNode"

    # AND the attribute value should reference the subworkflow node's output
    value = input_value_attr.get("value")
    assert isinstance(value, dict)
    assert value.get("type") == "NODE_OUTPUT"
    assert value.get("node_id") == subworkflow_node.get("id")

    # AND the node_output_id should match the actual output ID from the subworkflow
    actual_output_id = value.get("node_output_id")
    assert actual_output_id == expected_output_id, (
        f"node_output_id mismatch: expected {expected_output_id}, got {actual_output_id}. "
        "The reference to the subworkflow output is broken."
    )



from typing import TYPE_CHECKING, Type

from vellum.workflows.events.types import VellumCodeResourceDefinition

if TYPE_CHECKING:
    from vellum.workflows.events.workflow import WorkflowEvent
    from vellum.workflows.workflows.base import BaseWorkflow


def workflow_event_filter(workflow_definition: Type["BaseWorkflow"], event: "WorkflowEvent") -> bool:
    """
    Filters for only Workflow events that were emitted by the `workflow_definition` parameter.
    """

    if (
        event.name == "workflow.execution.initiated"
        or event.name == "workflow.execution.resumed"
        or event.name == "workflow.execution.fulfilled"
        or event.name == "workflow.execution.rejected"
        or event.name == "workflow.execution.paused"
        or event.name == "workflow.execution.streaming"
    ):
        return event.workflow_definition == workflow_definition

    return False


def root_workflow_event_filter(workflow_definition: Type["BaseWorkflow"], event: "WorkflowEvent") -> bool:
    """
    Filters for Workflow and Node events that were emitted by the `workflow_definition` parameter.
    """

    if (
        event.name == "workflow.execution.initiated"
        or event.name == "workflow.execution.resumed"
        or event.name == "workflow.execution.fulfilled"
        or event.name == "workflow.execution.rejected"
        or event.name == "workflow.execution.paused"
        or event.name == "workflow.execution.streaming"
    ):
        return event.workflow_definition == workflow_definition

    if not event.parent:
        return False

    if event.parent.type != "WORKFLOW":
        return False

    event_parent_definition = event.parent.workflow_definition
    current_workflow_definition = VellumCodeResourceDefinition.encode(workflow_definition)

    return event_parent_definition.model_dump() == current_workflow_definition.model_dump()


def workflow_sandbox_event_filter(workflow_definition: Type["BaseWorkflow"], event: "WorkflowEvent") -> bool:
    """
    Filter designed for Workflow Sandbox interfaces: include all events except
    workflow.execution.snapshotted events from nested/subworkflows. Only allow
    snapshotted events when they belong to the root workflow definition.
    """

    if event.name == "workflow.execution.snapshotted":
        return event.workflow_definition == workflow_definition

    return True


def all_workflow_event_filter(workflow_definition: Type["BaseWorkflow"], event: "WorkflowEvent") -> bool:
    return True



from uuid import UUID

from vellum_ee.workflows.display.editor import NodeDisplayData, NodeDisplayPosition
from vellum_ee.workflows.display.nodes import BaseTemplatingNodeDisplay
from vellum_ee.workflows.display.nodes.types import NodeOutputDisplay, PortDisplayOverrides

from ...nodes.templating_node import TemplatingNode


class TemplatingNodeDisplay(BaseTemplatingNodeDisplay[TemplatingNode]):
    label = "Templating Node"
    node_id = UUID("533c6bd8-6088-4abc-a168-8c1758abcd33")
    target_handle_id = UUID("048d310e-1a4d-4038-b534-8c68a4509b93")
    template_input_id = UUID("f97d721a-e685-498e-90c3-9c3d9358fdad")
    node_input_ids_by_name = {
        "example_var_1": UUID("a0d1d7cf-242a-4bd9-a437-d308a7ced9b3"),
        "template": UUID("f97d721a-e685-498e-90c3-9c3d9358fdad"),
    }
    output_display = {
        TemplatingNode.Outputs.result: NodeOutputDisplay(id=UUID("423bc529-1a1a-4f72-af4d-cbdb5f0a5929"), name="result")
    }
    port_displays = {
        TemplatingNode.Ports.default: PortDisplayOverrides(id=UUID("afda9a19-0618-42e1-9b63-5d0db2a88f62"))
    }
    display_data = NodeDisplayData(
        position=NodeDisplayPosition(x=1970.9393258426965, y=232.4943117977528), width=466, height=224
    )



"""Exceptions for Vellum file operations."""


class VellumFileError(Exception):
    """Base exception for all Vellum file errors."""

    pass


class InvalidFileSourceError(VellumFileError):
    """Raised when a file source URL/identifier is malformed or unsupported."""

    pass


class FileRetrievalError(VellumFileError):
    """Raised when a file cannot be retrieved from its source."""

    pass


class FileNotFoundError(FileRetrievalError):
    """Raised when a file cannot be found at its source."""

    pass



import pytest
from unittest.mock import MagicMock

from vellum.client.core.api_error import ApiError
from vellum.client.types.composio_execute_tool_response import ComposioExecuteToolResponse
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.state import BaseState
from vellum.workflows.state.base import StateMeta

from ..node import WebSearchNode


@pytest.fixture
def base_node_setup(vellum_client):
    """Basic node setup with required inputs."""

    class Inputs(BaseInputs):
        query: str

    class State(BaseState):
        pass

    class TestableWebSearchNode(WebSearchNode):
        query = Inputs.query

    state = State(meta=StateMeta(workflow_inputs=Inputs(query="test query")))
    context = MagicMock()
    context.vellum_client = vellum_client
    node = TestableWebSearchNode(state=state, context=context)
    return node


def test_successful_search_with_results(base_node_setup):
    """Test successful SerpAPI search with typical results."""

    # GIVEN a mock Composio SerpAPI response with results
    mock_response_data = {
        "results": {
            "organic_results": [
                {
                    "title": "First Result",
                    "snippet": "This is the first search result snippet",
                    "link": "https://example1.com",
                    "position": 1,
                },
                {
                    "title": "Second Result",
                    "link": "https://example2.com",
                    "position": 2,
                },
                {
                    "title": "Third Result",
                    "link": "https://example3.com",
                    "position": 3,
                },
            ]
        }
    }

    mock_response = ComposioExecuteToolResponse(provider="COMPOSIO", data=mock_response_data)
    base_node_setup._context.vellum_client.integrations.execute_integration_tool = MagicMock(return_value=mock_response)

    # WHEN we run the node
    outputs = base_node_setup.run()

    # THEN the text output should be properly formatted with labels
    expected_text = """\
Title: First Result
Snippet: This is the first search result snippet
URL: https://example1.com

Title: Second Result
URL: https://example2.com

Title: Third Result
URL: https://example3.com\
"""
    assert outputs.text == expected_text

    # AND URLs should be extracted correctly
    assert outputs.urls == ["https://example1.com", "https://example2.com", "https://example3.com"]

    # AND raw results should be serialized WebSearchResult models
    assert len(outputs.results) == 3
    assert outputs.results == [
        {
            "title": "First Result",
            "snippet": "This is the first search result snippet",
            "url": "https://example1.com",
        },
        {
            "title": "Second Result",
            "snippet": None,
            "url": "https://example2.com",
        },
        {
            "title": "Third Result",
            "snippet": None,
            "url": "https://example3.com",
        },
    ]

    # AND the Vellum client should have been called with the correct parameters
    base_node_setup._context.vellum_client.integrations.execute_integration_tool.assert_called_once_with(
        integration_name="SERPAPI",
        integration_provider="COMPOSIO",
        tool_name="SERPAPI_SEARCH",
        arguments={"query": "test query"},
    )


def test_authentication_error_401(base_node_setup):
    """Test authentication error raises NodeException with PROVIDER_CREDENTIALS_UNAVAILABLE."""

    base_node_setup._context.vellum_client.integrations.execute_integration_tool = MagicMock(
        side_effect=ApiError(status_code=403, body={"detail": "You must authenticate with the search provider."})
    )

    # WHEN we run the node
    with pytest.raises(NodeException) as exc_info:
        base_node_setup.run()

    # THEN it should raise the appropriate error with correct code
    assert exc_info.value.code == WorkflowErrorCode.PROVIDER_CREDENTIALS_UNAVAILABLE
    assert exc_info.value.message == "You must authenticate with the search provider."


def test_serpapi_error_in_response(base_node_setup):
    """Test SerpAPI error field in response raises NodeException."""

    # GIVEN SerpAPI returns an error in the response data
    mock_response = ComposioExecuteToolResponse(provider="COMPOSIO", data={"error": "Invalid search parameters"})
    base_node_setup._context.vellum_client.integrations.execute_integration_tool = MagicMock(return_value=mock_response)

    # WHEN we run the node
    with pytest.raises(NodeException) as exc_info:
        base_node_setup.run()

    # THEN it should raise the appropriate error
    assert exc_info.value.code == WorkflowErrorCode.PROVIDER_ERROR
    assert exc_info.value.message == "SerpAPI error: Invalid search parameters"


def test_empty_query_validation(vellum_client):
    """Test empty query raises validation error."""

    # GIVEN a node with an empty query
    class TestNode(WebSearchNode):
        query = ""

    context = MagicMock()
    context.vellum_client = vellum_client
    node = TestNode(state=BaseState(meta=StateMeta(workflow_inputs=BaseInputs())), context=context)

    # WHEN we run the node
    with pytest.raises(NodeException) as exc_info:
        node.run()

    # THEN it should raise a validation error
    assert exc_info.value.code == WorkflowErrorCode.INVALID_INPUTS
    assert exc_info.value.message == "Query is required and must be a non-empty string"


def test_empty_organic_results(base_node_setup):
    """Test handling of empty search results."""
    # GIVEN Composio SerpAPI returns no organic results in new format
    mock_response = ComposioExecuteToolResponse(provider="COMPOSIO", data={"results": {"organic_results": []}})
    base_node_setup._context.vellum_client.integrations.execute_integration_tool = MagicMock(return_value=mock_response)

    # WHEN we run the node
    outputs = base_node_setup.run()

    # THEN all outputs should be empty
    assert outputs.text == ""
    assert outputs.urls == []
    assert outputs.results == []


def test_missing_fields_in_results(base_node_setup):
    """Test handling of results with missing optional snippet field."""
    # GIVEN Composio SerpAPI returns results with some missing snippet fields
    mock_response_data = {
        "results": {
            "organic_results": [
                {
                    "title": "Result with snippet",
                    "link": "https://example1.com",
                    "snippet": "This has a snippet",
                },
                {
                    "title": "Result without snippet",
                    "link": "https://example2.com",
                },
                {
                    "title": "Another with snippet",
                    "link": "https://example3.com",
                    "snippet": "Another snippet here",
                },
            ]
        }
    }

    mock_response = ComposioExecuteToolResponse(provider="COMPOSIO", data=mock_response_data)
    base_node_setup._context.vellum_client.integrations.execute_integration_tool = MagicMock(return_value=mock_response)

    # WHEN we run the node
    outputs = base_node_setup.run()

    # THEN text should include all results with snippet only when present
    expected_text = """\
Title: Result with snippet
Snippet: This has a snippet
URL: https://example1.com

Title: Result without snippet
URL: https://example2.com

Title: Another with snippet
Snippet: Another snippet here
URL: https://example3.com\
"""
    assert outputs.text == expected_text

    # AND URLs should include all links
    assert outputs.urls == ["https://example1.com", "https://example2.com", "https://example3.com"]

    assert len(outputs.results) == 3
    assert outputs.results == [
        {
            "title": "Result with snippet",
            "snippet": "This has a snippet",
            "url": "https://example1.com",
        },
        {
            "title": "Result without snippet",
            "snippet": None,
            "url": "https://example2.com",
        },
        {
            "title": "Another with snippet",
            "snippet": "Another snippet here",
            "url": "https://example3.com",
        },
    ]



from datetime import datetime
from uuid import uuid4

from deepdiff import DeepDiff

from vellum import DeploymentRead
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_text_prompt_deployment.workflow import BasicTextPromptDeployment
from tests.workflows.basic_text_prompt_deployment.workflow_with_prompt_deployment_json_reference import (
    WorkflowWithPromptDeploymentJsonReferenceWorkflow,
)


def test_serialize_workflow(vellum_client):
    # GIVEN a Workflow with stubbed out API calls
    deployment = DeploymentRead(
        id=str(uuid4()),
        created=datetime.now(),
        label="Example Prompt Deployment",
        name="example_prompt_deployment",
        last_deployed_on=datetime.now(),
        input_variables=[],
        active_model_version_ids=[],
        last_deployed_history_item_id=str(uuid4()),
    )
    vellum_client.deployments.retrieve.return_value = deployment

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicTextPromptDeployment)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 2
    assert not DeepDiff(
        [
            {
                "id": "52995b50-84c9-465f-8a4b-a4ee2a92e388",
                "key": "city",
                "type": "STRING",
                "required": True,
                "default": None,
                "extensions": {"color": None},
            },
            {
                "id": "aa3ca842-250c-4a3f-853f-23928c28d0f8",
                "key": "date",
                "type": "STRING",
                "required": True,
                "default": None,
                "extensions": {"color": None},
            },
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert output_variables == [
        {
            "id": "a609ab19-db1b-4cd0-bdb0-aee5ed31dc28",
            "key": "text",
            "type": "STRING",
        }
    ]

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["edges"]) == 2
    assert len(workflow_raw_data["nodes"]) == 3

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "d680afbd-de64-4cf6-aa50-912686c48c64",
        "type": "ENTRYPOINT",
        "inputs": [],
        "base": None,
        "definition": None,
        "data": {
            "label": "Entrypoint Node",
            "source_handle_id": "7065a943-1cab-4afd-9690-e678c5b74a2f",
        },
        "display_data": {
            "position": {"x": 0.0, "y": -50.0},
        },
    }

    prompt_node = workflow_raw_data["nodes"][1]
    assert prompt_node == {
        "id": "a114abf2-cb76-49b5-8001-6c8df56d39ff",
        "type": "PROMPT",
        "inputs": [
            {
                "id": "c58d2244-a52e-4edb-84c5-efc334dea6f9",
                "key": "city",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "52995b50-84c9-465f-8a4b-a4ee2a92e388"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "d00707fb-b584-40bc-aecd-db11490bea3e",
                "key": "date",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "aa3ca842-250c-4a3f-853f-23928c28d0f8"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Example Prompt Deployment Node",
            "output_id": "f07b3521-4ba1-4a3c-8629-3a269406f519",
            "error_output_id": None,
            "array_output_id": "14d71e78-94a5-4c46-bad8-bec827a4f1e4",
            "source_handle_id": "6699f465-dc6c-4fa7-8038-7ff49419b953",
            "target_handle_id": "1407e51f-cb29-4a86-beeb-cc1870dc5525",
            "variant": "DEPLOYMENT",
            "prompt_deployment_id": deployment.id,
            "release_tag": "LATEST",
            "ml_model_fallbacks": ["gpt-4o", "gemini-1.5-pro"],
        },
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {
            "name": "PromptDeploymentNode",
            "module": ["vellum", "workflows", "nodes", "displayable", "prompt_deployment_node", "node"],
        },
        "definition": {
            "name": "ExamplePromptDeploymentNode",
            "module": ["tests", "workflows", "basic_text_prompt_deployment", "workflow"],
        },
        "trigger": {
            "id": "1407e51f-cb29-4a86-beeb-cc1870dc5525",
            "merge_behavior": "AWAIT_ANY",
        },
        "ports": [{"id": "6699f465-dc6c-4fa7-8038-7ff49419b953", "name": "default", "type": "DEFAULT"}],
        "outputs": [
            {"id": "66687de8-3e00-4290-9177-54be727fef44", "name": "json", "type": "JSON", "value": None},
            {"id": "f07b3521-4ba1-4a3c-8629-3a269406f519", "name": "text", "type": "STRING", "value": None},
            {"id": "14d71e78-94a5-4c46-bad8-bec827a4f1e4", "name": "results", "type": "ARRAY", "value": None},
        ],
    }

    final_output_node = workflow_raw_data["nodes"][2]
    assert final_output_node == {
        "id": "64ff72c7-8ffc-4e1f-b7a7-e7cd0697f576",
        "type": "TERMINAL",
        "base": {
            "module": [
                "vellum",
                "workflows",
                "nodes",
                "displayable",
                "final_output_node",
                "node",
            ],
            "name": "FinalOutputNode",
        },
        "definition": None,
        "inputs": [
            {
                "id": "78aeb65b-3491-4d2a-8c47-401d4cb3d560",
                "key": "node_input",
                "value": {
                    "combinator": "OR",
                    "rules": [
                        {
                            "data": {
                                "node_id": "a114abf2-cb76-49b5-8001-6c8df56d39ff",
                                "output_id": "f07b3521-4ba1-4a3c-8629-3a269406f519",
                            },
                            "type": "NODE_OUTPUT",
                        }
                    ],
                },
            }
        ],
        "data": {
            "label": "Final Output",
            "name": "text",
            "node_input_id": "78aeb65b-3491-4d2a-8c47-401d4cb3d560",
            "output_id": "a609ab19-db1b-4cd0-bdb0-aee5ed31dc28",
            "output_type": "STRING",
            "target_handle_id": "dced939a-9122-4290-8482-7daa9525dad6",
        },
        "display_data": {
            "position": {
                "x": 400.0,
                "y": -50.0,
            },
        },
    }

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert serialized_edges == [
        {
            "id": "89a6042d-2bfd-4512-9f75-bf11db3c493c",
            "source_handle_id": "7065a943-1cab-4afd-9690-e678c5b74a2f",
            "source_node_id": "d680afbd-de64-4cf6-aa50-912686c48c64",
            "target_handle_id": "1407e51f-cb29-4a86-beeb-cc1870dc5525",
            "target_node_id": "a114abf2-cb76-49b5-8001-6c8df56d39ff",
            "type": "DEFAULT",
        },
        {
            "id": "c2cbf6ef-8582-45c8-a643-fc6ae8fe482f",
            "source_handle_id": "6699f465-dc6c-4fa7-8038-7ff49419b953",
            "source_node_id": "a114abf2-cb76-49b5-8001-6c8df56d39ff",
            "target_handle_id": "dced939a-9122-4290-8482-7daa9525dad6",
            "target_node_id": "64ff72c7-8ffc-4e1f-b7a7-e7cd0697f576",
            "type": "DEFAULT",
        },
    ]

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "BasicTextPromptDeployment",
        "module": [
            "tests",
            "workflows",
            "basic_text_prompt_deployment",
            "workflow",
        ],
    }


def test_serialize_workflow_with_prompt_and_templating(vellum_client):
    # GIVEN a Workflow with stubbed out API calls
    deployment = DeploymentRead(
        id=str(uuid4()),
        created=datetime.now(),
        label="JSON Prompt Deployment",
        name="json_prompt_deployment",
        last_deployed_on=datetime.now(),
        input_variables=[],
        active_model_version_ids=[],
        last_deployed_history_item_id=str(uuid4()),
    )
    vellum_client.deployments.retrieve.return_value = deployment

    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=WorkflowWithPromptDeploymentJsonReferenceWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 2
    assert not DeepDiff(
        [
            {
                "id": "ad577aca-0797-4deb-a484-574ef1a1f0c7",
                "key": "city",
                "type": "STRING",
                "default": None,
                "required": True,
                "extensions": {"color": None},
            },
            {
                "id": "066124c4-42bd-4764-aa75-6f230dbbed4a",
                "key": "date",
                "type": "STRING",
                "default": None,
                "required": True,
                "extensions": {"color": None},
            },
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 1
    assert output_variables == [{"id": "a7e4b449-5879-4d0c-8f00-d5d4985eb65c", "key": "text", "type": "STRING"}]

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["edges"]) == 3
    assert len(workflow_raw_data["nodes"]) == 4

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "32c7f398-277c-456b-9279-aa1f867fb637",
        "type": "ENTRYPOINT",
        "inputs": [],
        "data": {"label": "Entrypoint Node", "source_handle_id": "cc0f4028-1039-4063-971d-7dacbb01b379"},
        "display_data": {"position": {"x": 0.0, "y": -50.0}},
        "base": None,
        "definition": None,
    }

    prompt_node = workflow_raw_data["nodes"][1]
    assert prompt_node == {
        "id": "0d1460e4-f207-4a69-bcea-7a3c7b325c02",
        "type": "PROMPT",
        "inputs": [
            {
                "id": "b99dec61-8c38-4b6b-96fa-cb83d5c9a9ef",
                "key": "city",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "ad577aca-0797-4deb-a484-574ef1a1f0c7"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "6379398c-3c29-4327-9346-4d386a467f16",
                "key": "date",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "066124c4-42bd-4764-aa75-6f230dbbed4a"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Example Prompt Deployment Node",
            "output_id": "b9d579e2-b3ee-4524-917f-7329fc09af59",
            "error_output_id": None,
            "array_output_id": "547b4a00-eb16-4df5-92f4-cdd2fe7a0848",
            "source_handle_id": "7e29137d-af96-402c-8108-9a00e087d18e",
            "target_handle_id": "f8017ad7-14f2-4e6f-8456-a081db5ed7cd",
            "variant": "DEPLOYMENT",
            "prompt_deployment_id": deployment.id,
            "release_tag": "LATEST",
            "ml_model_fallbacks": None,
        },
        "display_data": {"position": {"x": 200.0, "y": -50.0}},
        "base": {
            "name": "PromptDeploymentNode",
            "module": ["vellum", "workflows", "nodes", "displayable", "prompt_deployment_node", "node"],
        },
        "definition": {
            "name": "ExamplePromptDeploymentNode",
            "module": [
                "tests",
                "workflows",
                "basic_text_prompt_deployment",
                "workflow_with_prompt_deployment_json_reference",
            ],
        },
        "trigger": {
            "id": "f8017ad7-14f2-4e6f-8456-a081db5ed7cd",
            "merge_behavior": "AWAIT_ANY",
        },
        "ports": [{"id": "7e29137d-af96-402c-8108-9a00e087d18e", "name": "default", "type": "DEFAULT"}],
        "outputs": [
            {"id": "62bbe13a-9571-4165-9463-22092a04e450", "name": "json", "type": "JSON", "value": None},
            {"id": "b9d579e2-b3ee-4524-917f-7329fc09af59", "name": "text", "type": "STRING", "value": None},
            {"id": "547b4a00-eb16-4df5-92f4-cdd2fe7a0848", "name": "results", "type": "ARRAY", "value": None},
        ],
    }

    templating_node = workflow_raw_data["nodes"][2]
    assert templating_node == {
        "id": "ac2de275-729f-4fba-9701-97beba80df34",
        "type": "TEMPLATING",
        "inputs": [
            {
                "id": "198f6350-237a-4103-b571-311738b7743f",
                "key": "template",
                "value": {
                    "rules": [
                        {
                            "type": "CONSTANT_VALUE",
                            "data": {
                                "type": "STRING",
                                "value": "The weather in {{ city }} on {{ date }} is {{ weather }}.",
                            },
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "2439a747-3b8d-44e9-85c1-99b903d69997",
                "key": "city",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "ad577aca-0797-4deb-a484-574ef1a1f0c7"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "d85dfd04-f304-4bbf-9d24-52d9ef8f45cc",
                "key": "date",
                "value": {
                    "rules": [
                        {
                            "type": "INPUT_VARIABLE",
                            "data": {"input_variable_id": "066124c4-42bd-4764-aa75-6f230dbbed4a"},
                        }
                    ],
                    "combinator": "OR",
                },
            },
            {
                "id": "a555cbb1-cea9-4454-add0-61609b38a376",
                "key": "weather",
                "value": {
                    "rules": [
                        {
                            "type": "NODE_OUTPUT",
                            "data": {
                                "node_id": "0d1460e4-f207-4a69-bcea-7a3c7b325c02",
                                "output_id": "62bbe13a-9571-4165-9463-22092a04e450",
                            },
                        }
                    ],
                    "combinator": "OR",
                },
            },
        ],
        "data": {
            "label": "Example Templating Node",
            "output_id": "c8ad13db-d601-4c24-bef7-0196baa8079c",
            "error_output_id": None,
            "source_handle_id": "bfe059fb-987f-47dd-bfbb-c71fdf1e4971",
            "target_handle_id": "fbd5aa14-b615-42c9-a85a-23eb1a6b5436",
            "template_node_input_id": "198f6350-237a-4103-b571-311738b7743f",
            "output_type": "STRING",
        },
        "display_data": {"position": {"x": 400.0, "y": -50.0}},
        "base": {
            "name": "TemplatingNode",
            "module": ["vellum", "workflows", "nodes", "core", "templating_node", "node"],
        },
        "definition": {
            "name": "ExampleTemplatingNode",
            "module": [
                "tests",
                "workflows",
                "basic_text_prompt_deployment",
                "workflow_with_prompt_deployment_json_reference",
            ],
        },
        "trigger": {
            "id": "fbd5aa14-b615-42c9-a85a-23eb1a6b5436",
            "merge_behavior": "AWAIT_ATTRIBUTES",
        },
        "ports": [{"id": "bfe059fb-987f-47dd-bfbb-c71fdf1e4971", "name": "default", "type": "DEFAULT"}],
    }

    final_output_node = workflow_raw_data["nodes"][3]
    assert final_output_node == {
        "id": "53de824d-a41d-4294-b511-c969932b05af",
        "type": "TERMINAL",
        "data": {
            "label": "Final Output",
            "name": "text",
            "target_handle_id": "fee3d395-38c3-485f-ab61-1a0fdf71c4ce",
            "output_id": "a7e4b449-5879-4d0c-8f00-d5d4985eb65c",
            "output_type": "STRING",
            "node_input_id": "cf380f81-c5ee-4bc9-8e26-ecf1307733a9",
        },
        "inputs": [
            {
                "id": "cf380f81-c5ee-4bc9-8e26-ecf1307733a9",
                "key": "node_input",
                "value": {
                    "rules": [
                        {
                            "type": "NODE_OUTPUT",
                            "data": {
                                "node_id": "ac2de275-729f-4fba-9701-97beba80df34",
                                "output_id": "c8ad13db-d601-4c24-bef7-0196baa8079c",
                            },
                        }
                    ],
                    "combinator": "OR",
                },
            }
        ],
        "display_data": {"position": {"x": 600.0, "y": -50.0}},
        "base": {
            "name": "FinalOutputNode",
            "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
        },
        "definition": None,
    }

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert serialized_edges == [
        {
            "id": "33e0220a-969a-4d38-86ac-0fe596ab2d88",
            "source_node_id": "32c7f398-277c-456b-9279-aa1f867fb637",
            "source_handle_id": "cc0f4028-1039-4063-971d-7dacbb01b379",
            "target_node_id": "0d1460e4-f207-4a69-bcea-7a3c7b325c02",
            "target_handle_id": "f8017ad7-14f2-4e6f-8456-a081db5ed7cd",
            "type": "DEFAULT",
        },
        {
            "id": "8eb009df-4daf-4ff6-8dc2-ff71a2829e2b",
            "source_node_id": "0d1460e4-f207-4a69-bcea-7a3c7b325c02",
            "source_handle_id": "7e29137d-af96-402c-8108-9a00e087d18e",
            "target_node_id": "ac2de275-729f-4fba-9701-97beba80df34",
            "target_handle_id": "fbd5aa14-b615-42c9-a85a-23eb1a6b5436",
            "type": "DEFAULT",
        },
        {
            "id": "1f720900-e5e1-49b7-9910-6ede79f6afd2",
            "source_node_id": "ac2de275-729f-4fba-9701-97beba80df34",
            "source_handle_id": "bfe059fb-987f-47dd-bfbb-c71fdf1e4971",
            "target_node_id": "53de824d-a41d-4294-b511-c969932b05af",
            "target_handle_id": "fee3d395-38c3-485f-ab61-1a0fdf71c4ce",
            "type": "DEFAULT",
        },
    ]

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "WorkflowWithPromptDeploymentJsonReferenceWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_text_prompt_deployment",
            "workflow_with_prompt_deployment_json_reference",
        ],
    }
